# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of vLLM copy_blocks_kernel."""

from typing import Final

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _copy_blocks_kernel(
    # Pointers to tensors
    key_cache_ptrs: tl.tensor,
    value_cache_ptrs: tl.tensor,
    block_mapping_ptr: tl.const,
    # Strides of relevant tensors
    kv_cache_block_stride: int,
    block_mapping_pair_stride: int,
    # Scalars
    page_size: int,
    # Constexprs
    kv_cache_dtype: tl.constexpr,
    cxpr_chunk_size: tl.constexpr,
) -> None:
    """Implementation of copy_blocks kernel.

    Args:
        key_cache_ptrs: Tensor where each element is a pointer to a key cache block, shape: (num_layers,); cache_shape: (num_blocks, page_size).
        value_cache_ptrs: Tensor where each element is a pointer to a value cache block, shape: (num_layers,); cache_shape: (num_blocks, page_size).
        block_mapping_ptr: Tensor listing source/destination pairs to copy, shape: (num_pairs, 2).
        kv_cache_block_stride: Stride of key/value cache tensors in the 0th dimension.
        block_mapping_pair_stride: Stride of block mapping tensor in the 0th dimension.
        page_size: Size of a page (set of cache blocks).
        kv_cache_dtype: Data type of KV cache.
        cxpr_chunk_size: Size of chunk of memory to process at once.
    """
    # What layer of the cache are we processing?
    layer_index = tl.program_id(0)
    # What pair of src/dst blocks are we copying?
    pair_index = tl.program_id(1)

    # Consider adding additional program axis here to parallelize across number of chunks.
    # Could choose to loop or spawn more programs depending on num_chunks.

    # Map constexpr cache dtype to tl.dtype (cannot pass directly to kernel w/o constexpr and cannot use constexpr as argument to tl.pointer_type())
    triton_kv_cache_dtype = tl.float32
    if kv_cache_dtype == tl.float16:
        triton_kv_cache_dtype = tl.float16
    if kv_cache_dtype == tl.bfloat16:
        triton_kv_cache_dtype = tl.bfloat16
    if kv_cache_dtype == tl.uint8:
        triton_kv_cache_dtype = tl.uint8

    # Get pointers to the key/value cache tensors for this layer
    layer_key_cache_ptr = tl.load(key_cache_ptrs + layer_index).to(tl.pointer_type(triton_kv_cache_dtype))
    layer_value_cache_ptr = tl.load(value_cache_ptrs + layer_index).to(tl.pointer_type(triton_kv_cache_dtype))

    # Calculate offset to pair of src/dst block ids
    pair_offset = pair_index * block_mapping_pair_stride

    # Load the source/destination block ids
    source_block = tl.load(block_mapping_ptr + pair_offset)
    destination_block = tl.load(block_mapping_ptr + pair_offset + 1)

    # Calculate offset from the start of the k/v caches to the start of the source/destination blocks
    source_block_offset = source_block * kv_cache_block_stride
    destination_block_offset = destination_block * kv_cache_block_stride

    # We may not have enough memory to store the entire page in SRAM, so we copy in sizes of {cxpr_chunk_size}
    num_chunks = tl.cdiv(page_size, cxpr_chunk_size)
    for chunk_index in tl.range(0, num_chunks):
        # Get offsets from the start of the k/v caches for this chunk of the page
        chunk_offsets = (chunk_index * cxpr_chunk_size) + tl.arange(0, cxpr_chunk_size)
        # Handle out-of-bounds accesses
        chunk_mask = chunk_offsets < page_size

        # Load the source key block
        source_key_block = tl.load(layer_key_cache_ptr + source_block_offset + chunk_offsets, mask=chunk_mask)
        # Store it at the destination key block
        tl.store(layer_key_cache_ptr + destination_block_offset + chunk_offsets, source_key_block, mask=chunk_mask)

        # Load the source value block
        source_value_block = tl.load(layer_value_cache_ptr + source_block_offset + chunk_offsets, mask=chunk_mask)
        # Store it at the destination value block
        tl.store(layer_value_cache_ptr + destination_block_offset + chunk_offsets, source_value_block, mask=chunk_mask)


def _torch_to_triton_dtype(torch_dtype: torch.dtype) -> tl.dtype:
    """Convert Torch dtype to corresponding Triton dtype.

    Args:
        torch_dtype: Torch datatype.

    Raises:
        ValueError: if cannot match Torch dtype to corresponding Triton dtype.

    Returns:
        Corresponding Triton datatype.
    """
    if torch_dtype in (torch.float32, torch.float):
        return tl.float32

    if torch_dtype in (torch.float16, torch.half):
        return tl.float16

    if torch_dtype == torch.bfloat16:
        return tl.bfloat16

    if torch_dtype == torch.uint8:
        return tl.uint8

    msg = f"Cannot map Torch dtype ({torch_dtype}) to Triton dtype"
    raise ValueError(msg)


def copy_blocks_launcher(
    key_caches: list[torch.Tensor],
    value_caches: list[torch.Tensor],
    block_mapping: torch.Tensor,
) -> None:
    """Launch copy_blocks kernel.

    Args:
        key_caches: List of key_cache for a set of layers, len: num_layers, each element shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        value_caches: List of value_cache for a set of layers, len: num_layers, num_layers, each element shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        block_mapping: Tensor in form of list of source/destination pairs, shape: (num_pairs, 2)
    """
    # Assume sizes already checked if calling launcher. For interface with strict size checking, call `copy_blocks()`.
    num_layers: Final = len(key_caches)
    num_pairs: Final = block_mapping.size(0)
    device: Final = key_caches[0].device
    kv_cache_dtype: Final = key_caches[0].dtype
    kv_cache_block_stride: Final = key_caches[0].stride(0)
    block_mapping_pair_stride: Final = block_mapping.stride(0)
    page_size: Final = key_caches[0].size(1)

    key_cache_ptrs = [key_cache_tensor.data_ptr() for key_cache_tensor in key_caches]
    key_cache_ptrs_tensor = torch.tensor(key_cache_ptrs, device=device, dtype=torch.uint64)

    value_cache_ptrs = [value_cache_tensor.data_ptr() for value_cache_tensor in value_caches]
    value_cache_ptrs_tensor = torch.tensor(value_cache_ptrs, device=device, dtype=torch.uint64)

    triton_kv_cache_dtype: Final = _torch_to_triton_dtype(kv_cache_dtype)

    cxpr_chunk_size: tl.constexpr = 1024

    # Parallelize over each layer and each pair of copies
    grid = (num_layers, num_pairs)

    # Launch kernel
    _copy_blocks_kernel[grid](
        key_cache_ptrs_tensor,
        value_cache_ptrs_tensor,
        block_mapping,
        kv_cache_block_stride,
        block_mapping_pair_stride,
        page_size,
        triton_kv_cache_dtype,
        cxpr_chunk_size,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch paged attention benchmark."""

import sys
from typing import Final

import click
import torch

from conch import envs
from conch.ops.attention.paged_attention import paged_attention
from conch.platforms import current_platform
from conch.third_party.vllm.utils import create_tensors, seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it

if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
    from vllm._custom_ops import paged_attention_v2 as vllm_paged_attention_v2
else:
    vllm_paged_attention_v2 = None  # type: ignore[assignment, unused-ignore]


@click.command()
@click.option(
    "--head-dim",
    required=True,
    type=int,
    default=128,
    help="Head dimension",
)
@click.option(
    "--seq-len",
    required=True,
    type=int,
    default=8192,
    help="Sequence length (for k/v)",
)
@click.option(
    "--cache-block-size",
    required=True,
    type=int,
    default=32,
    help="Number of KV vectors in each cache block",
)
@click.option(
    "--batch-size",
    required=False,
    type=int,
    default=128,
    help="Batch size",
)
@click.option(
    "--num-query-heads",
    required=False,
    type=int,
    default=32,
    help="Number of query heads",
)
@click.option(
    "--num-kv-heads",
    required=False,
    type=int,
    default=8,
    help="Number of kv heads",
)
@click.option(
    "--kv-cache-dtype",
    required=False,
    type=click.Choice(["auto", "fp8", "fp8_e4m3"]),
    default="auto",
    help="Dtype of KV cache",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    head_dim: int,
    seq_len: int,
    cache_block_size: int,
    batch_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    kv_cache_dtype: str,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark PagedAttention.

    Args:
        head_dim: Head dimension of input tensors.
        seq_len: Sequence length of input tensors.
        cache_block_size: Number of KV vectors in each cache block.
        batch_size: Batch size of input tensors.
        num_query_heads: Number of attention query heads.
        num_kv_heads: Number of attention kv heads.
        kv_cache_dtype: Data type of the kv cache.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag to indicate whether or not to print results in CSV format.
    """
    if kv_cache_dtype != "auto" and not current_platform.supports_fp8():
        error_msg = f"kv_cache_type '{kv_cache_dtype}' not supported on this GPU!"
        raise NotImplementedError(error_msg)

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = torch.float16

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_dim": head_dim,
            "seq_len": seq_len,
            "cache_block_size": cache_block_size,
            "batch_size": batch_size,
            "num_query_heads": num_query_heads,
            "num_kv_heads": num_kv_heads,
        },
    )

    query, key_cache_vllm, value_cache_vllm, key_cache_conch, value_cache_conch, block_table, seq_lens = create_tensors(
        head_dim,
        seq_len,
        cache_block_size,
        batch_size,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        device,
        dtype,
    )

    scale: Final = float(1.0 / (head_dim**0.5))
    max_seq_len: Final = int(seq_lens.max().item())

    k_scale = torch.full((1,), 0.5, dtype=dtype, device=device)
    v_scale = torch.full((1,), 0.5, dtype=dtype, device=device)

    output_conch = torch.empty_like(query)

    paged_attention(
        query,
        key_cache_conch,
        value_cache_conch,
        block_table,
        seq_lens,
        output=output_conch,
        scale=scale,
        softcap=0.0,
        kv_cache_dtype=kv_cache_dtype,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    if vllm_paged_attention_v2 is not None:
        # VLLM scratchpads
        partition_size: Final = 512
        max_num_partitions = (max_seq_len + partition_size - 1) // partition_size
        tmp_output = torch.empty(
            size=(batch_size, num_query_heads, max_num_partitions, head_dim),
            dtype=query.dtype,
            device=query.device,
        )
        exp_sums = torch.empty(
            size=(batch_size, num_query_heads, max_num_partitions),
            dtype=torch.float32,
            device=query.device,
        )
        max_logits = torch.empty_like(exp_sums)

        alibi_slopes = None

        output_vllm = torch.empty_like(query)

        # Check accuracy match
        vllm_paged_attention_v2(
            output_vllm,
            exp_sums,
            max_logits,
            tmp_output,
            query,
            key_cache_vllm,
            value_cache_vllm,
            num_kv_heads,
            scale,
            block_table,
            seq_lens,
            cache_block_size,
            max_seq_len,
            alibi_slopes,
            kv_cache_dtype,
            k_scale,
            v_scale,
        )

        if not torch.allclose(output_vllm, output_conch, atol=absolute_tolerance):
            print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
            print(f"Output max diff: {(output_conch - output_vllm).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Reference output: {output_vllm}", file=sys.stderr)
                print(f"Conch output: {output_conch}", file=sys.stderr)
        else:
            print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

        baseline_result = benchmark_it(
            lambda: vllm_paged_attention_v2(
                output_vllm,
                exp_sums,
                max_logits,
                tmp_output,
                query,
                key_cache_vllm,
                value_cache_vllm,
                num_kv_heads,
                scale,
                block_table,
                seq_lens,
                cache_block_size,
                max_seq_len,
                alibi_slopes,
                kv_cache_dtype,
                k_scale,
                v_scale,
            ),
            tag="Baseline",
            metadata=metadata,
            iteration_time_ms=iteration_time_ms,
            warmup_time_ms=warmup_time_ms,
        )
    else:
        print("Skipping checking vs. reference vLLM implementation...", file=sys.stderr)
        baseline_result = None

    conch_result = benchmark_it(
        lambda: paged_attention(
            query,
            key_cache_conch,
            value_cache_conch,
            block_table,
            seq_lens,
            output=output_conch,
            scale=scale,
            softcap=0.0,
            kv_cache_dtype=kv_cache_dtype,
            k_scale=k_scale,
            v_scale=v_scale,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    if baseline_result is not None:
        baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""PyTorch reference implementation of vLLM reshape_and_cache."""

import torch

from conch import envs
from conch.platforms import current_platform


def _reshape_and_cache_pytorch_ref(
    key: torch.Tensor,
    value: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    slot_mapping: torch.Tensor,
    kv_cache_dtype: str,
    k_scale: torch.Tensor,
    v_scale: torch.Tensor,
) -> None:
    """Reference PyTorch-only implementation of reshape_and_cache."""
    num_tokens, _, _ = key.shape
    _, _, _, block_size, _ = key_cache.shape

    if kv_cache_dtype == "fp8":
        k_scale_scalar = 1.0 / k_scale.item()
        v_scale_scalar = 1.0 / v_scale.item()
        fp8_dtype = torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn

    reshaped_key = key.reshape(num_tokens, *key_cache[0, :, :, 0, :].shape)
    block_indicies = torch.div(slot_mapping, block_size, rounding_mode="floor")
    block_indicies_lst = block_indicies.cpu().tolist()
    block_offsets = slot_mapping % block_size
    block_offsets_lst = block_offsets.cpu().tolist()
    for i in range(num_tokens):
        block_idx = block_indicies_lst[i]
        block_offset = block_offsets_lst[i]
        if kv_cache_dtype == "fp8":
            key_cache[block_idx, :, :, block_offset, :] = (reshaped_key[i] * k_scale_scalar).to(fp8_dtype)
            value_cache[block_idx, :, :, block_offset] = (value[i] * v_scale_scalar).to(fp8_dtype)
        else:
            key_cache[block_idx, :, :, block_offset, :] = reshaped_key[i]
            value_cache[block_idx, :, :, block_offset] = value[i]


def _reshape_and_cache_vllm_ref(
    key: torch.Tensor,
    value: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    slot_mapping: torch.Tensor,
    kv_cache_dtype: str,
    k_scale: torch.Tensor,
    v_scale: torch.Tensor,
) -> None:
    """Reference vLLM implementation of reshape_and_cache."""
    from vllm._custom_ops import reshape_and_cache as reshape_and_cache_vllm

    reshape_and_cache_vllm(key, value, key_cache, value_cache, slot_mapping, kv_cache_dtype, k_scale, v_scale)


def reshape_and_cache(
    key: torch.Tensor,
    value: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    slot_mapping: torch.Tensor,
    kv_cache_dtype: str,
    k_scale: torch.Tensor,
    v_scale: torch.Tensor,
) -> None:
    """Reference implementation of vLLM's reshape_and_cache operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _reshape_and_cache_vllm_ref(
            key,
            value,
            key_cache,
            value_cache,
            slot_mapping,
            kv_cache_dtype=kv_cache_dtype,
            k_scale=k_scale,
            v_scale=v_scale,
        )

    return _reshape_and_cache_pytorch_ref(
        key,
        value,
        key_cache,
        value_cache,
        slot_mapping,
        kv_cache_dtype=kv_cache_dtype,
        k_scale=k_scale,
        v_scale=v_scale,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of bitsandbytes quantization."""

from enum import Enum

import torch
import triton
import triton.language as tl


class QuantizedDType(Enum):
    """Data types for mapping between Triton/Torch."""

    GENERAL_8BIT = 0
    NF4 = 1
    FP4 = 2


# Triton can only access global values instantiated as tl.constexpr
_QUANTIZED_DTYPE_GENERAL_8BIT: tl.constexpr = tl.constexpr(QuantizedDType.GENERAL_8BIT.value)
_QUANTIZED_DTYPE_NF4: tl.constexpr = tl.constexpr(QuantizedDType.NF4.value)
_QUANTIZED_DTYPE_FP4: tl.constexpr = tl.constexpr(QuantizedDType.FP4.value)


def _get_quantized_dtype(quant_type: str) -> QuantizedDType:
    """Map string quant_type to enum."""
    if quant_type == "nf4":
        return QuantizedDType.NF4
    if quant_type == "fp4":
        return QuantizedDType.FP4
    return QuantizedDType.GENERAL_8BIT


@triton.jit  # type: ignore[misc]
def _nf4_quantize(x: tl.tensor) -> int:  # noqa: C901, PLR0911, PLR0912
    """Quantize a floating point value to NF4."""
    if x > 0.8614784181118011:  # noqa: PLR2004
        return 15
    if x > 0.6427869200706482:  # noqa: PLR2004
        return 14
    if x > 0.5016634166240692:  # noqa: PLR2004
        return 13
    if x > 0.3893125355243683:  # noqa: PLR2004
        return 12
    if x > 0.2920137718319893:  # noqa: PLR2004
        return 11
    if x > 0.2035212516784668:  # noqa: PLR2004
        return 10
    if x > 0.1202552504837513:  # noqa: PLR2004
        return 9
    if x > 0.03979014977812767:  # noqa: PLR2004
        return 8
    if x > -0.045525018125772476:  # noqa: PLR2004
        return 7
    if x > -0.13791173323988914:  # noqa: PLR2004
        return 6
    if x > -0.23460740596055984:  # noqa: PLR2004
        return 5
    if x > -0.33967943489551544:  # noqa: PLR2004
        return 4
    if x > -0.4599952697753906:  # noqa: PLR2004
        return 3
    if x > -0.6106329262256622:  # noqa: PLR2004
        return 2
    if x > -0.8480964004993439:  # noqa: PLR2004
        return 1
    return 0


@triton.jit  # type: ignore[misc]
def _fp4_quantize(x: tl.tensor) -> int:  # noqa: C901, PLR0911, PLR0912
    """Quantize a floating point value to FP4."""
    sign = 8 if x < 0 else 0
    x = tl.abs(x)

    if x > 0.29166667:
        if x > 0.5833334:
            if x > 0.83333334:
                return 3 + sign
            return 2 + sign
        if x > 0.4166667:
            return 5 + sign
        return 4 + sign

    if x > 0.0859375:
        if x > 0.208333334:
            return 7 + sign
        return 6 + sign
    if x > 0.00260417:
        return 1 + sign
    return 0 + sign


@triton.jit  # type: ignore[misc]
def _fp8_quantize(x: tl.tensor, code_ptr: tl.tensor) -> int:
    """Quantize a floating point value to FP8 (bitcast-ed to uint8)."""
    pivot = 127
    upper_pivot = 255
    lower_pivot = 0

    lower = -1.0
    upper = 1.0

    val = tl.load(code_ptr + pivot)

    # Iterate backwards through powers of two: [64, 32, 16, 8, 4, 2, 1]
    for current_power_of_two in range(6, -1, -1):
        i = tl.exp2(current_power_of_two.to(tl.float32)).to(tl.int32)  # type: ignore[attr-defined]

        if x > val:
            lower_pivot = pivot
            lower = val
            pivot += i
        else:
            upper_pivot = pivot
            upper = val
            pivot -= i

        val = tl.load(code_ptr + pivot)

    if upper_pivot == 255:  # noqa: PLR2004
        upper = tl.load(code_ptr + upper_pivot)
    if lower_pivot == 0:
        lower = tl.load(code_ptr + lower_pivot)

    if x > val:
        midpoint = (upper + val) * 0.5
        if x > midpoint:
            return upper_pivot
        return pivot

    midpoint = (lower + val) * 0.5
    if x < midpoint:
        return lower_pivot
    return pivot


@triton.jit  # type: ignore[misc]
def _quantize_blockwise_kernel(  # noqa: PLR0913
    x_ptr: tl.tensor,
    absmax_ptr: tl.tensor,
    out_ptr: tl.tensor,
    code_ptr: tl.tensor,
    input_size: int,
    cxpr_blocksize: tl.constexpr,
    cxpr_output_blocksize: tl.constexpr,
    cxpr_quantized_dtype: tl.constexpr,
) -> None:
    """Quantize blockwise kernel."""
    block_index = tl.program_id(0)

    input_block_offset = block_index * cxpr_blocksize

    block_offsets = input_block_offset + tl.arange(0, cxpr_blocksize)
    block_mask = block_offsets < input_size

    # Load this block of the input
    block = tl.load(x_ptr + block_offsets, mask=block_mask, other=0.0)

    # Reduce absmax and store for de-quantization
    local_absmax = tl.max(tl.abs(block))
    tl.store(absmax_ptr + block_index, local_absmax)

    # Invert so that we can _multiply_ by scaling factor during quantization
    local_absmax = 1.0 / local_absmax.to(tl.float32)

    # Don't read extra elements if input_size is not perfectly divisible by blocksize
    output_block_offset = block_index * cxpr_output_blocksize
    this_block_size = min(cxpr_blocksize, input_size - input_block_offset)
    this_output_block_size = (
        this_block_size if cxpr_quantized_dtype == _QUANTIZED_DTYPE_GENERAL_8BIT else this_block_size // 2
    )

    for element_index in range(this_output_block_size):
        if cxpr_quantized_dtype == _QUANTIZED_DTYPE_GENERAL_8BIT:
            x = tl.load(x_ptr + input_block_offset + element_index)
            result = _fp8_quantize(x.to(tl.float32) * local_absmax, code_ptr)
            tl.store(out_ptr + output_block_offset + element_index, result)
        else:
            # We quantize two values with each iteration of this loop
            packed_element_offset = element_index * 2
            # Load two elements (note: unfortunately we cannot index into these elements even though we already loaded them)
            x1 = tl.load(x_ptr + input_block_offset + packed_element_offset)
            x2 = tl.load(x_ptr + input_block_offset + packed_element_offset + 1)

            # Cast to fp32, scale by inverse of max(abs()) of block
            x1 = x1.to(tl.float32) * local_absmax
            x2 = x2.to(tl.float32) * local_absmax

            # Initialize result to 0 (i.e. bitwise-or always flips the bit)
            result = 0

            # Apply quantization and bitshift
            if cxpr_quantized_dtype == _QUANTIZED_DTYPE_FP4:
                result |= _fp4_quantize(x1) << 4
                result |= _fp4_quantize(x2)
            else:
                result |= _nf4_quantize(x1) << 4
                result |= _nf4_quantize(x2)

            # Store
            tl.store(out_ptr + output_block_offset + element_index, result)


def quantize_blockwise_launcher(  # noqa: PLR0913
    x: torch.Tensor,
    absmax: torch.Tensor,
    out: torch.Tensor,
    code: torch.Tensor | None,
    blocksize: int,
    input_size: int,
    quant_type: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Convert FP32/FP16/BF16 input tensor to FP4/NF4/FP8."""
    # Blocksize must be a power-of-two
    assert blocksize == triton.next_power_of_2(blocksize)  # noqa: S101
    if quant_type in ("fp4", "nf4"):
        # Only support even input size if packing two elements into one
        assert input_size % 2 == 0  # noqa: S101

    quantized_dtype = _get_quantized_dtype(quant_type)
    if quantized_dtype == QuantizedDType.GENERAL_8BIT:
        assert code is not None  # noqa: S101

    # If 4-bit, we're packing two elements from the input into each element of the output
    output_blocksize = blocksize if quantized_dtype == QuantizedDType.GENERAL_8BIT else blocksize // 2

    # How many blocks are we processing?
    num_blocks = triton.cdiv(input_size, blocksize)

    # Process each block separately
    grid = (num_blocks,)

    # Launch kernel
    _quantize_blockwise_kernel[grid](
        x_ptr=x,
        absmax_ptr=absmax,
        # Need output to be treated as uint8
        # See: https://huggingface.co/docs/bitsandbytes/main/en/fsdp_qlora#quantized-data-storage
        out_ptr=out.view(torch.uint8),
        code_ptr=code,
        input_size=input_size,
        cxpr_blocksize=blocksize,
        cxpr_output_blocksize=output_blocksize,
        cxpr_quantized_dtype=quantized_dtype.value,
    )

    return out, absmax



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test for silu_and_mul Triton port."""

from typing import Final

import pytest
import torch

from conch.ops.activation.silu_and_mul import silu_and_mul as silu_and_mul_triton
from conch.platforms import current_platform
from conch.reference.activation.silu_and_mul import silu_and_mul as silu_and_mul_reference
from conch.third_party.vllm.utils import seed_everything

_DTYPES: Final = [torch.float32, torch.float16, torch.bfloat16]
_DIM: Final = [855, 2048]
_BATCH_SIZE: Final = [4, 10, 32]
_NUM_TOKENS: Final = [20, 512]
_USE_BATCH_SIZE: Final = [True, False]


@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("d", _DIM)
@pytest.mark.parametrize("batch_size", _BATCH_SIZE)
@pytest.mark.parametrize("num_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("use_batch_size", _USE_BATCH_SIZE)
def test_silu_and_mul(dtype: torch.dtype, d: int, batch_size: int, num_tokens: int, use_batch_size: bool) -> None:
    """Test Triton SiLU implementation."""
    seed_everything(0)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    x_shape = (batch_size, num_tokens, 2 * d) if use_batch_size else (num_tokens, 2 * d)
    x = torch.randn(x_shape, dtype=dtype, device=device)

    ref_output = silu_and_mul_reference(x)
    triton_output = silu_and_mul_triton(x)

    tolerance = 1e-3
    torch.testing.assert_close(triton_output, ref_output, rtol=tolerance, atol=tolerance)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Bitsandbytes dequantize PyTorch reference implementation."""

from collections.abc import Callable

import torch
import triton


def _pytorch_fp4_dequantize(x: torch.Tensor, code: torch.Tensor | None) -> float:  # noqa: C901, PLR0911, PLR0912
    """Dequantize an FP4 value to floating point."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is None

    if x == 15:
        return -0.25
    if x == 14:
        return -0.166666
    if x == 13:
        return -0.5
    if x == 12:
        return -0.333333
    if x == 11:
        return -1.0
    if x == 10:
        return -0.666666
    if x == 9:
        return -0.0052083333
    if x == 8:
        return -0.0
    if x == 7:
        return 0.25
    if x == 6:
        return 0.166666
    if x == 5:
        return 0.5
    if x == 4:
        return 0.333333
    if x == 3:
        return 1.0
    if x == 2:
        return 0.6666666
    if x == 1:
        return 0.0052083333
    return 0.0


def _pytorch_nf4_dequantize(x: torch.Tensor, code: torch.Tensor | None) -> float:  # noqa: C901, PLR0911, PLR0912
    """Dequantize an NF4 value to floating point."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is None
    if x == 15:
        return 1.0
    if x == 14:
        return 0.7229568362236023
    if x == 13:
        return 0.5626170039176941
    if x == 12:
        return 0.44070982933044434
    if x == 11:
        return 0.33791524171829224
    if x == 10:
        return 0.24611230194568634
    if x == 9:
        return 0.16093020141124725
    if x == 8:
        return 0.07958029955625534
    if x == 7:
        return 0.0
    if x == 6:
        return -0.09105003625154495
    if x == 5:
        return -0.18477343022823334
    if x == 4:
        return -0.28444138169288635
    if x == 3:
        return -0.39491748809814453
    if x == 2:
        return -0.5250730514526367
    if x == 1:
        return -0.6961928009986877
    return -1.0


def _pytorch_fp8_dequantize(x: torch.Tensor, code: torch.Tensor | None) -> float:
    """Dequantize an FP8 value to fp32."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is not None
    return code[int(x.item())].item()


def _get_pytorch_dequant_method(quant_type: str) -> Callable[[torch.Tensor, torch.Tensor | None], float]:
    """Get PyTorch dequantization method."""
    if quant_type == "nf4":
        return _pytorch_nf4_dequantize
    if quant_type == "fp4":
        return _pytorch_fp4_dequantize
    return _pytorch_fp8_dequantize


def dequantize_blockwise_launcher(  # noqa: PLR0913
    x: torch.Tensor,
    absmax: torch.Tensor,
    out: torch.Tensor,
    code: torch.Tensor | None,
    blocksize: int,
    output_size: int,
    quant_type: str,
) -> torch.Tensor:
    """Convert FP32/FP16/BF16 input tensor to NF4/FP4/FP8."""
    assert x.dtype == torch.uint8

    dequant_method = _get_pytorch_dequant_method(quant_type)

    num_blocks = triton.cdiv(output_size, blocksize)
    input_blocksize = blocksize // 2 if quant_type in ("fp4", "nf4") else blocksize

    for block_index in range(num_blocks):
        input_block_offset = block_index * input_blocksize
        output_block_offset = block_index * blocksize

        local_absmax = absmax[block_index]

        this_output_block_size = min(blocksize, output_size - output_block_offset)
        this_input_block_size = this_output_block_size // 2 if quant_type in ("fp4", "nf4") else this_output_block_size

        for element_index in range(this_input_block_size):
            if quant_type in ("fp4", "nf4"):
                packed = x[input_block_offset + element_index]
                x1 = dequant_method(packed >> 4, code) * local_absmax
                x2 = dequant_method(packed & 0x0F, code) * local_absmax

                output_element_offset = element_index * 2
                out[output_block_offset + output_element_offset] = x1
                out[output_block_offset + output_element_offset + 1] = x2
            else:
                result = dequant_method(x[input_block_offset + element_index], code) * local_absmax
                out[output_block_offset + element_index] = result

    return out



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for Triton reshape_and_cache."""

import random
from typing import Final

import pytest
import torch

from conch.ops.vllm.reshape_and_cache import reshape_and_cache as reshape_and_cache_triton
from conch.platforms import current_platform
from conch.reference.vllm.reshape_and_cache import reshape_and_cache as reshape_and_cache_reference
from conch.third_party.vllm.utils import create_kv_caches_with_random, reshape_vllm_kvcache, seed_everything

_DTYPES: Final = [torch.float16, torch.bfloat16, torch.float32]
_NUM_TOKENS: Final = [20, 40, 60]
_NUM_HEADS: Final = [1, 4]
_HEAD_SIZES: Final = [128]
_BLOCK_SIZES: Final = [32, 128]
_NUM_BLOCKS: Final = [1000]
_KV_CACHE_DTYPE: Final = ["auto", "fp8"]


@pytest.mark.parametrize("num_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("num_heads", _NUM_HEADS)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize("block_size", _BLOCK_SIZES)
@pytest.mark.parametrize("num_blocks", _NUM_BLOCKS)
@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("kv_cache_dtype", _KV_CACHE_DTYPE)
@torch.inference_mode()
def test_reshape_and_cache(
    num_tokens: int,
    num_heads: int,
    head_size: int,
    block_size: int,
    num_blocks: int,
    dtype: torch.dtype,
    kv_cache_dtype: str,
) -> None:
    """Test Triton reshape_and_cache vs. reference implementations."""
    if kv_cache_dtype != "auto" and not current_platform.supports_fp8():
        pytest.skip("FP8 is not supported on this GPU type.")

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    # Create a random slot mapping.
    num_slots = block_size * num_blocks
    slot_mapping_lst = random.sample(range(num_slots), num_tokens)
    slot_mapping = torch.tensor(slot_mapping_lst, dtype=torch.long)

    kv = torch.randn((num_tokens, 2, num_heads, head_size), dtype=dtype)
    key, value = kv.unbind(dim=1)

    k_scale = torch.full((1,), 2.0, dtype=torch.float32, device=device)
    v_scale = torch.full((1,), 3.0, dtype=torch.float32, device=device)

    # Create the KV caches.
    key_caches_vllm, value_caches_vllm = create_kv_caches_with_random(
        num_blocks,
        block_size,
        1,
        num_heads,
        head_size,
        kv_cache_dtype,
        dtype,
        seed,
        device,
    )

    key_cache_vllm, value_cache_vllm = key_caches_vllm[0], value_caches_vllm[0]

    if "fp8" in kv_cache_dtype:
        fp8_dtype = torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn
        key_cache_vllm = key_cache_vllm.view(fp8_dtype)
        value_cache_vllm = value_cache_vllm.view(fp8_dtype)

    key_cache_conch, value_cache_conch = reshape_vllm_kvcache(key_cache_vllm, value_cache_vllm)

    # Run the reference implementation.
    reshape_and_cache_reference(
        key, value, key_cache_vllm, value_cache_vllm, slot_mapping, kv_cache_dtype, k_scale, v_scale
    )

    # Call Triton kernel
    reshape_and_cache_triton(
        key, value, key_cache_conch, value_cache_conch, slot_mapping, kv_cache_dtype, k_scale, v_scale
    )

    # Reshape vLLM key/value caches
    key_cache_vllm, value_cache_vllm = reshape_vllm_kvcache(key_cache_vllm, value_cache_vllm)

    # Can't compare FP8 directly, so bitcast to uint8
    if "fp8" in kv_cache_dtype:
        key_cache_vllm = key_cache_vllm.view(torch.uint8)
        value_cache_vllm = value_cache_vllm.view(torch.uint8)
        key_cache_conch = key_cache_conch.view(torch.uint8)
        value_cache_conch = value_cache_conch.view(torch.uint8)

    # Compare the results.
    torch.testing.assert_close(key_cache_conch, key_cache_vllm)
    torch.testing.assert_close(value_cache_conch, value_cache_vllm)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Port of vllm gelu_tanh_and_mul to Triton."""

from typing import Final

import torch
import triton
import triton.language as tl
from triton.language.extra import libdevice  # type: ignore[attr-defined]

_M_2_SQRTPI: Final = 1.12837916709551257390
_M_SQRT2: Final = 1.41421356237309504880


@triton.jit  # type: ignore[misc]
def _gelu_tanh_and_mul_kernel(
    out_ptr: tl.tensor,  # [..., d]
    out_stride: int,
    in_ptr: tl.tensor,  # [..., 2 * d]
    in_stride: int,
    d: int,
    cxpr_block_size: tl.constexpr,
    cxpr_beta: tl.constexpr,
    cxpr_kappa: tl.constexpr,
) -> None:
    """Apply tanh-approximated GeLU and multiply.

    This step is part of the GeGLU activation used in the GEMMA model.
    https://arxiv.org/abs/2002.05202
    https://storage.googleapis.com/gweb-developer-goog-blog-assets/images/image2_l7UnOuC.original.png

    vllm refs:
    vllm::act_and_mul_kernel (wraps user-defined ACT and performs mul)
    vllm::gelu_tanh_kernel
    vllm::gelu_tanh_and_mul
    LAUNCH_ACTIVATION_GATE_KERNEL

    Ref implementation summary:
    - One block per token
    - Threads work in parallel to perform point-wise GeLU and multiply
    - Threads will loop until all d elements are processed

    Args:
        out_ptr: Storage for results; expected dimensions (num_tokens, projection_dim).
        out_stride: step size between output rows.
        in_ptr: Adjoined projections to be scaled [ ..., [gate_proj | up_proj]]; ([opt-batch], num_tokens, projection_dim * 2).
        in_stride: step size between input rows.
        d: dimensionality of projection space.
        cxpr_block_size: working block size.
        cxpr_beta: BETA constant.
        cxpr_kappa: KAPPA constant.
    """
    token_idx = tl.program_id(axis=0)
    in_block_start = in_ptr + (token_idx * in_stride)
    local_offset = tl.arange(0, cxpr_block_size)
    out_block_start = out_ptr + (token_idx * out_stride)
    for _ in tl.range(0, d, cxpr_block_size):
        mask = local_offset < d
        x = tl.load(in_block_start + local_offset, mask=mask)
        y = tl.load(in_block_start + d + local_offset, mask=mask)
        x_cubed = x * x * x
        inner = cxpr_beta * (x + cxpr_kappa * x_cubed)
        gelu_act = 0.5 * x * (1 + libdevice.tanh(inner))
        scaled = gelu_act * y
        tl.store(out_block_start + local_offset, scaled, mask=mask)
        local_offset += cxpr_block_size


def gelu_tanh_and_mul_launcher(
    output: torch.Tensor,
    projections: torch.Tensor,
) -> None:
    """gelu_tanh_and_mul launcher.

    Args:
        output: Storage for results; expected dimensions (num_tokens, projection_dim)
        projections: Adjoined projections to be scaled; (num_tokens, projection_dim * 2)
    """
    d = projections.shape[-1] // 2
    num_tokens = projections.numel() // (2 * d)
    beta = _M_SQRT2 * _M_2_SQRTPI * 0.5
    kappa = 0.044715
    block_size = min(256, triton.next_power_of_2(d))

    # Note about using negative indices in shape and stride
    # The commonly (?) accepted organization of data appears to be:
    # - batch, token, other higher level dimensions in the leading positions
    # - data-specific dimensions in the lower positions
    # For example: [batch_size, num_token, embedding_size]
    # Some higher level dimensions may or may not exist.  For example, for unbatched data, there wouldn't be a
    # batch_size dimension.  On the other hand, the lower-level dimensions will always exist.  This is why it makes
    # some sense to count backwards when trying to access dimension information.

    _gelu_tanh_and_mul_kernel[(num_tokens,)](
        output,
        output.stride(-2),
        projections,
        projections.stride(-2),
        d,
        cxpr_block_size=block_size,
        cxpr_beta=beta,
        cxpr_kappa=kappa,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test Triton port of GemmaRMSNorm."""

from typing import Final

import pytest
import torch

from conch.ops.normalization.gemma_rms_norm import gemma_rms_norm as gemma_rms_norm_triton
from conch.platforms import current_platform
from conch.reference.normalization.gemma_rms_norm import gemma_rms_norm as gemma_rms_norm_reference
from conch.third_party.vllm.utils import seed_everything

_N_POS: Final = [2, 500, 1000]
_HIDDEN_SIZES: Final = [1024, 2048, 4096]


@pytest.mark.parametrize("n_pos", _N_POS)
@pytest.mark.parametrize("hidden_size", _HIDDEN_SIZES)
@torch.inference_mode()
def test_kernel(
    n_pos: int,
    hidden_size: int,
) -> None:
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    epsilon: Final = 1e-6

    x = torch.randn((n_pos, hidden_size), dtype=torch.float16, device=device)
    weights = torch.randn((hidden_size,), device=device)

    x_ref = x.clone()
    x_triton = x.clone()

    result_ref = gemma_rms_norm_reference(x_ref, weights, epsilon, residual=None)
    result_triton = gemma_rms_norm_triton(x_triton, weights, epsilon, residual=None)

    torch.testing.assert_close(result_ref, result_triton, atol=1e-5, rtol=0.001)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Platform enum."""

import enum
from dataclasses import dataclass
from typing import Final

import torch


class PlatformEnum(enum.Enum):
    """Enum for different platforms."""

    NVIDIA = enum.auto()
    AMD = enum.auto()
    XPU = enum.auto()
    CPU = enum.auto()
    UNSPECIFIED = enum.auto()


@dataclass
class Platform:
    """Class representing a platform."""

    platform_enum: PlatformEnum
    device: str

    def name(self) -> str:
        """Return the name of the platform."""
        return self.platform_enum.name

    def is_nvidia(self) -> bool:
        """Check if the platform is NVIDIA."""
        return self.platform_enum == PlatformEnum.NVIDIA

    def is_amd(self) -> bool:
        """Check if the platform is AMD."""
        return self.platform_enum == PlatformEnum.AMD

    def is_unspecified(self) -> bool:
        """Check if the platform is unspecified."""
        return self.platform_enum == PlatformEnum.UNSPECIFIED

    def has_cuda(self) -> bool:
        """Check if the platform has CUDA support."""
        return self.is_nvidia() or self.is_amd()

    def supports_fp8(self) -> bool:
        """Check if the platform supports FP8."""
        if torch.cuda.is_available():
            # Triton: fp8e4nv data type is not supported on CUDA arch < 89
            min_fp8_capability_triton: Final = 89

            capability = torch.cuda.get_device_capability()
            actual_capability = capability[0] * 10 + capability[1]
            return actual_capability >= min_fp8_capability_triton

        # Until we officially support XPU/CPU, just assume True so that we can uncover any issues
        return True

    def get_device_name(self) -> str:
        if torch.cuda.is_available():
            return torch.cuda.get_device_name()

        return "unknown"


def detect_current_platform() -> Platform:
    """Detect the current platform."""
    # We could do much more sophisticated things to detect the platform,
    # but this is an easy first-cut
    if hasattr(torch, "cuda") and torch.cuda.is_available():
        if torch.version.cuda is not None:
            return Platform(PlatformEnum.NVIDIA, "cuda")

        if torch.version.hip is not None:
            return Platform(PlatformEnum.AMD, "cuda")

    if hasattr(torch, "xpu") and torch.xpu.is_available():
        return Platform(PlatformEnum.XPU, "xpu")

    if hasattr(torch, "cpu") and torch.cpu.is_available():
        return Platform(PlatformEnum.CPU, "cpu")

    return Platform(PlatformEnum.UNSPECIFIED, "cpu")



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test for rotary_embedding."""

from typing import Final

import pytest
import torch

from conch.ops.embedding.rotary_embedding import rotary_embedding as rotary_embedding_triton
from conch.platforms import current_platform
from conch.reference.embedding.rotary_embedding import compute_cos_sin_cache
from conch.reference.embedding.rotary_embedding import rotary_embedding as rotary_embedding_reference
from conch.third_party.vllm.utils import seed_everything

_SEQ_LENS: Final = [2048, 8192]
_NUM_HEADS: Final = [4, 17]
_HEAD_SIZES: Final = [128, 256]
_BASES: Final = [10000]


@pytest.mark.parametrize("seq_len", _SEQ_LENS)
@pytest.mark.parametrize("num_heads", _NUM_HEADS)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize("base", _BASES)
@torch.inference_mode()
def test_kernel(
    seq_len: int,
    num_heads: int,
    head_size: int,
    base: int,
) -> None:
    rotary_dim = head_size
    max_position = seq_len
    is_neox_style = True

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    cos_sin_cache = compute_cos_sin_cache(base, rotary_dim, max_position)

    # seq_len == num_tokens in this case
    positions = torch.randint(0, max_position, (seq_len,))
    query = torch.randn(seq_len, num_heads * head_size)
    key = torch.randn_like(query)
    query_ref = torch.clone(query)  # Need to clone since updates are performed in-place
    key_ref = torch.clone(key)

    query_ref, key_ref = rotary_embedding_reference(
        positions,
        query_ref,
        key_ref,
        cos_sin_cache,
        rotary_dim,
        head_size,
        is_neox_style=is_neox_style,
    )

    query_triton, key_triton = rotary_embedding_triton(
        positions,
        query,
        key,
        head_size,
        cos_sin_cache,
        is_neox=is_neox_style,
    )

    torch.testing.assert_close(query_ref, query)
    torch.testing.assert_close(key_ref, key)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Create markdown table for README."""

import os
from pathlib import Path
from subprocess import run
from typing import Final

import click
import pandas as pd  # type: ignore[import-untyped]

from conch.platforms import current_platform

_MARKDOWN_TABLE_HEADER: Final = """\
| Operation | CUDA Runtime | Triton Runtime | Triton Speedup |
| --- | --- | --- | --- |
"""

_MARKDOWN_TABLE_ROW: Final = "| {} | {:.3f} ms | {:.3f} ms | {:.2f} |\n"

_TABLE_OP_NAME_TO_BENCHMARK: Final = {
    "GeLU, Tanh, and Mul": "gelu_tanh_and_mul_benchmark",
    "SiLU and Mul": "silu_and_mul_benchmark",
    "Paged Attention": "paged_attention_vs_flash_benchmark",
    "Varlen Attention": "varlen_attention_benchmark",
    "Rotary Embedding": "rotary_embedding_benchmark",
    "RMS Norm (Gemma-style)": "gemma_rms_norm_benchmark",
    "RMS Norm (Llama-style)": "rms_norm_benchmark",
    "bitsandbytes: Dequantize": "bnb_dequantize_blockwise_benchmark",
    "bitsandbytes: Quantize": "bnb_quantize_blockwise_benchmark",
    "FP8 Static Quantization": "static_scaled_fp8_quant_benchmark",
    "Int8 Static Quantization": "static_scaled_int8_quant_benchmark",
    "Mixed-precision GEMM [Int4 x FP16]": "mixed_precision_gemm_benchmark",
    "Scaled GEMM [Int8 x BF16]": "scaled_gemm_benchmark",
    "vLLM: Copy Blocks": "copy_blocks_benchmark",
    "vLLM: Reshape and Cache": "reshape_and_cache_benchmark",
}

_DEVICE_SPECIFIC_BLACKLIST: Final = {
    "NVIDIA A10": [
        "static_scaled_fp8_quant_benchmark",
        "mixed_precision_gemm_benchmark",
    ],
    "unknown": [],
}

# Add any extra flags for each benchmark here
_EXTRA_BENCHMARK_FLAGS: Final = {
    "varlen_attention_benchmark": ["--causal"],
}


@click.command()
@click.option(
    "--results-directory",
    required=True,
    type=click.Path(path_type=Path, dir_okay=True, file_okay=False),
    help="Path to the benchmark results directory.",
)
@click.option(
    "--use-cached-results",
    is_flag=True,
    help="Flag to use cached benchmark results if they exist",
)
def main(results_directory: Path, use_cached_results: bool) -> None:
    """Main function to plot benchmarking results."""
    # Always run against fastest possible implementation
    os.environ["CONCH_BENCH_ENABLE_ALL_REF"] = "1"
    os.environ["CONCH_ENABLE_BNB"] = "1"
    os.environ["CONCH_ENABLE_VLLM"] = "1"
    os.environ["VLLM_LOGGING_LEVEL"] = "CRITICAL"

    # Create directory for output if it doesn't exist already
    results_directory.mkdir(parents=True, exist_ok=True)

    # Running string for markdown table output
    result = _MARKDOWN_TABLE_HEADER

    device_name = current_platform.get_device_name()

    for op_name, benchmark_name in _TABLE_OP_NAME_TO_BENCHMARK.items():
        if (
            device_name in _DEVICE_SPECIFIC_BLACKLIST.keys()
            and benchmark_name in _DEVICE_SPECIFIC_BLACKLIST[device_name]
        ):
            print(f"Skipping {op_name} benchmark because it is blacklisted on current platform ({device_name = })")
            continue

        results_csv = results_directory / f"{benchmark_name}.csv"

        if use_cached_results and results_csv.exists():
            print(f"Skipping {op_name} benchmark, CSV file already exists.")
        else:
            # Run benchmark and redirect output
            print(f"Running benchmark for {op_name}...")

            # Some benchmark args are flags to enable things that default false, so we add any per-benchmark here
            extra_flags = _EXTRA_BENCHMARK_FLAGS[benchmark_name] if benchmark_name in _EXTRA_BENCHMARK_FLAGS else []

            with results_csv.open("w") as results_file:
                run(
                    ["python", f"benchmarks/{benchmark_name}.py", "--csv"] + extra_flags,
                    check=True,
                    stdout=results_file,
                    env=os.environ,
                )

        # Read the CSV file
        df = pd.read_csv(results_csv)
        triton_df = df[df["tag"] == "Conch"]
        baseline_df = df[df["tag"] == "Baseline"]

        # Calculate speedup
        triton_runtime = triton_df["runtime_ms"].iloc[0]
        baseline_runtime = baseline_df["runtime_ms"].iloc[0]
        speedup = baseline_runtime / triton_runtime

        # Create a markdown table row
        row = _MARKDOWN_TABLE_ROW.format(
            op_name,
            baseline_runtime,
            triton_runtime,
            speedup,
        )
        result += row

    # Print markdown table
    print("\nResults:\n")
    print(result)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Mixed-precision matrix multiplication kernel benchmark."""

import math
import sys
from typing import Final

import click
import torch

from conch import envs
from conch.ops.quantization.gemm import mixed_precision_gemm
from conch.platforms import current_platform
from conch.third_party.vllm.quant_utils import pack_rows, quantize_weights
from conch.third_party.vllm.scalar_type import ScalarType, scalar_types
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it

if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
    from vllm import _custom_ops as vllm_custom_ops
else:
    vllm_custom_ops = None  # type: ignore[assignment, unused-ignore]


def _to_torch_dtype(dtype_str: str) -> torch.dtype:
    """Map click arg for dtype to torch type."""
    if dtype_str == "bf16":
        return torch.bfloat16
    if dtype_str == "fp16":
        return torch.float16
    if dtype_str == "fp32":
        return torch.float32

    error_msg = f"Unrecognized data type: '{dtype_str}'"
    raise ValueError(error_msg)


def _to_scalar_dtype(dtype_str: str) -> ScalarType:
    """Map click arg for dtype to scalar type."""
    if dtype_str == "uint4":
        return scalar_types.uint4
    if dtype_str == "uint8":
        return scalar_types.uint8
    if dtype_str == "uint4b8":
        return scalar_types.uint4b8
    if dtype_str == "uint8b128":
        return scalar_types.uint8b128

    error_msg = f"Unrecognized data type: '{dtype_str}'"
    raise ValueError(error_msg)


def _machete_quantize_and_pack(
    w: torch.Tensor, wtype: ScalarType, group_size: int, enable_machete: bool
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Quantize and pack weight matrix."""
    w_ref, w_q, w_s, _ = quantize_weights(
        w,
        wtype,
        group_size,
        zero_points=False,
    )

    w_q_packed = pack_rows(w_q, wtype.size_bits, *w_q.shape)

    w_q_machete = w_q_packed.t().contiguous().t()
    if enable_machete and vllm_custom_ops is not None:
        w_q_machete = vllm_custom_ops.machete_prepack_B(
            w_q_machete, a_type=w.dtype, b_type=wtype, group_scales_type=w.dtype
        )

    return w_ref, w_q_machete, w_q_packed, w_s


@click.command()
@click.option(
    "--m-dim",
    required=True,
    type=int,
    default=4096,
    help="1st dimension of A matrix (M x K)",
)
@click.option(
    "--k-dim",
    required=True,
    type=int,
    default=8192,
    help="Common dimension of A and B matrices (M x K) * (K * N)",
)
@click.option(
    "--n-dim",
    required=True,
    type=int,
    default=4096,
    help="2nd dimension of B matrix (K x N)",
)
@click.option(
    "--input-dtype",
    required=True,
    type=click.Choice(["fp16", "bf16", "fp32"]),
    default="fp16",
    help="Data type of input",
)
@click.option(
    "--weight-dtype",
    required=True,
    type=click.Choice(["uint4", "uint8", "uint4b8", "uint8b128"]),
    default="uint4b8",
    help="Data type of input",
)
@click.option(
    "--enable-machete",
    is_flag=True,
    default=envs.CONCH_BENCH_ENABLE_ALL_REF,
    help="Flag for enabling running Machete (only on H100)",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    m_dim: int,
    k_dim: int,
    n_dim: int,
    input_dtype: str,
    weight_dtype: str,
    enable_machete: bool,
    iteration_time_ms: int,
    warmup_time_ms: int,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark mixed-precision matrix multiply.

    Args:
        m_dim: 1st dimension of matrix A.
        k_dim: Common dimension between matrix A and B.
        n_dim: 2nd dimension of matrix B.
        input_dtype: Data type of input matrices.
        weight_dtype: Data type of weight matrices.
        enable_machete: Enable running Machete kernel.
        iteration_time_ms: Time in milliseconds to run the benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    input_dtype_torch: Final = _to_torch_dtype(input_dtype)
    weight_dtype_vllm: Final = _to_scalar_dtype(weight_dtype)

    group_size: Final = 128
    assert group_size <= k_dim

    if enable_machete and vllm_custom_ops is None:
        error_msg = "In order to enable machete baseline we vLLM must be enabled via `CONCH_ENABLE_VLLM=1`."
        raise ValueError(error_msg)

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "m_dim": m_dim,
            "k_dim": k_dim,
            "n_dim": n_dim,
            "input_dtype": input_dtype,
            "weight_dtype": weight_dtype,
        },
    )

    a = (10 * (torch.rand((m_dim, k_dim), dtype=torch.float32, device=device) - 0.3)).to(input_dtype_torch)
    b = (10 * (torch.rand((k_dim, n_dim), dtype=torch.float32, device=device) - 0.3)).to(input_dtype_torch)

    w_ref, w_q_machete, w_q, w_s = _machete_quantize_and_pack(b, weight_dtype_vllm, group_size, enable_machete)

    output_ref = torch.matmul(a, w_ref)

    conch_output = mixed_precision_gemm(
        a, w_q, w_s, None, weight_dtype_vllm.size_bits, weight_dtype_vllm.bias, group_size
    )

    # Relax atol as our reduction dim becomes larger (more rounding error)
    atol = min(5e-2 * math.sqrt(k_dim), 1)
    rtol = 1e-1

    if enable_machete and vllm_custom_ops is not None:
        machete_output = vllm_custom_ops.machete_mm(
            a=a,
            b_q=w_q_machete,
            b_type=weight_dtype_vllm,
            b_group_scales=w_s,
            b_group_size=group_size,
        )

        if not torch.allclose(output_ref, machete_output, rtol=rtol, atol=atol):
            print("WARNING: Reference and machete results differ!", file=sys.stderr)
            print(f"Output max diff: {(output_ref - machete_output).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Reference output: {output_ref}", file=sys.stderr)
                print(f"Machete output: {machete_output}", file=sys.stderr)

    if not torch.allclose(output_ref, conch_output, rtol=rtol, atol=atol):
        print("WARNING: Reference and Conch results differ!", file=sys.stderr)
        print(f"Output max diff: {(output_ref - conch_output).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {output_ref}", file=sys.stderr)
            print(f"Conch output: {conch_output}", file=sys.stderr)
    else:
        print("Results matched :)", file=sys.stderr)

    if enable_machete and vllm_custom_ops is not None:
        baseline_result = benchmark_it(
            lambda: vllm_custom_ops.machete_mm(
                a=a,
                b_q=w_q_machete,
                b_type=weight_dtype_vllm,
                b_group_scales=w_s,
                b_group_size=group_size,
            ),
            tag="Baseline",
            metadata=metadata,
            iteration_time_ms=iteration_time_ms,
            warmup_time_ms=warmup_time_ms,
        )
    else:
        baseline_result = None

    conch_result = benchmark_it(
        lambda: mixed_precision_gemm(
            a, w_q, w_s, None, weight_dtype_vllm.size_bits, weight_dtype_vllm.bias, group_size
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    if baseline_result is not None:
        baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch varlen attention vs. FlashAttnVarlen benchmark."""

import sys
from typing import Final

import click
import torch

from conch import envs
from conch.ops.attention.varlen_attention import varlen_attention
from conch.platforms import current_platform
from conch.third_party.vllm.unified_attention import unified_attention
from conch.third_party.vllm.utils import create_tensors, seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it

if envs.CONCH_ENABLE_VLLM and current_platform.is_nvidia():
    from vllm.vllm_flash_attn import flash_attn_varlen_func  # type: ignore[attr-defined, unused-ignore]
else:
    flash_attn_varlen_func = None  # type: ignore[assignment, unused-ignore]


@click.command()
@click.option(
    "--head-dim",
    required=True,
    type=int,
    default=128,
    help="Head dimension",
)
@click.option(
    "--seq-len",
    required=True,
    type=int,
    default=512,
    help="Sequence length (for k/v)",
)
@click.option(
    "--cache-block-size",
    required=True,
    type=int,
    default=32,
    help="Number of KV vectors in each cache block",
)
@click.option(
    "--batch-size",
    required=False,
    type=int,
    default=64,
    help="Batch size",
)
@click.option(
    "--num-query-heads",
    required=False,
    type=int,
    default=32,
    help="Number of query heads",
)
@click.option(
    "--num-kv-heads",
    required=False,
    type=int,
    default=8,
    help="Number of kv heads",
)
@click.option(
    "--causal",
    is_flag=True,
    help="Flag to toggle causal/non-causal attention",
)
@click.option(
    "--pure-decode",
    is_flag=True,
    help="Flag for making all q_seqlens == 1",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    head_dim: int,
    seq_len: int,
    cache_block_size: int,
    batch_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    causal: bool,
    pure_decode: bool,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark VarlenAttention.

    Args:
        head_dim: Head dimension of input tensors.
        seq_len: Sequence length of input tensors.
        cache_block_size: Number of KV vectors in each cache block.
        batch_size: Batch size of input tensors.
        num_query_heads: Number of attention query heads.
        num_kv_heads: Number of attention kv heads.
        causal: Flag to toggle causal/non-causal attention.
        pure_decode: Flag for making all q_seqlens == 1.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag to indicate whether or not to print results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    scale: Final = float(1.0 / (head_dim**0.5))

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_dim": head_dim,
            "seq_len": seq_len,
            "cache_block_size": cache_block_size,
            "batch_size": batch_size,
            "num_query_heads": num_query_heads,
            "num_kv_heads": num_kv_heads,
            "causal": causal,
            "pure_decode": pure_decode,
        },
    )

    kv_cache_dtype: Final = "auto"
    dtype: Final = torch.float16

    _, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_dim,
        seq_len,
        cache_block_size,
        batch_size,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        current_platform.device,
        dtype,
    )

    starting_item = torch.as_tensor([0], dtype=torch.int32)

    if pure_decode:
        seqlens_q = torch.ones((batch_size,), dtype=torch.int32)

        cu_seqlens_q = torch.cumsum(seqlens_q, dim=0, dtype=torch.int32)
        cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

        max_seqlen_q = int(torch.max(seqlens_q).item())
        max_seqlen_k = int(torch.max(seq_lens).item())
    else:
        cu_seqlens_q = torch.cumsum(seq_lens, dim=0, dtype=torch.int32)

        cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

        max_seqlen_q = int(torch.max(seq_lens).item())
        max_seqlen_k = int(max_seqlen_q)

    total_num_q = int(cu_seqlens_q[-1].item())

    query = torch.empty((total_num_q, num_query_heads, head_dim), dtype=dtype, device=device)
    query.uniform_(-scale, scale)

    output_conch = varlen_attention(
        query=query,
        key_cache=key_cache,
        value_cache=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        seq_lens=seq_lens,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        scale=scale,
        causal=causal,
    )

    if flash_attn_varlen_func is not None:
        output_vllm = flash_attn_varlen_func(
            q=query,
            k=key_cache,
            v=value_cache,
            cu_seqlens_q=cu_seqlens_q,
            max_seqlen_q=max_seqlen_q,
            max_seqlen_k=max_seqlen_k,
            block_table=block_table,
            seqused_k=seq_lens,
            softmax_scale=scale,
            causal=causal,
        )

        if not torch.allclose(output_vllm, output_conch, atol=absolute_tolerance):
            print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
            print(f"Output max diff: {(output_conch - output_vllm).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Reference output: {output_vllm}", file=sys.stderr)
                print(f"Conch output: {output_conch}", file=sys.stderr)
        else:
            print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

        baseline_result = benchmark_it(
            lambda: flash_attn_varlen_func(
                q=query,
                k=key_cache,
                v=value_cache,
                cu_seqlens_q=cu_seqlens_q,
                max_seqlen_q=max_seqlen_q,
                max_seqlen_k=max_seqlen_k,
                block_table=block_table,
                seqused_k=seq_lens,
                softmax_scale=scale,
                causal=causal,
            ),
            tag="Baseline",
            metadata=metadata,
            iteration_time_ms=iteration_time_ms,
            warmup_time_ms=warmup_time_ms,
        )
    else:
        print("Skipping checking vs. reference vLLM implementation...", file=sys.stderr)
        baseline_result = None

    if causal and flash_attn_varlen_func is None:
        out = torch.zeros_like(query)

        vllm_triton_result = benchmark_it(
            lambda: unified_attention(
                q=query,
                k=key_cache,
                v=value_cache,
                out=out,
                cu_seqlens_q=cu_seqlens_q,
                max_seqlen_q=max_seqlen_q,
                seqused_k=seq_lens,
                max_seqlen_k=max_seqlen_k,
                softmax_scale=scale,
                causal=causal,
                window_size=(-1, -1),
                block_table=block_table,
                softcap=0.0,
                q_descale=None,
                k_descale=None,
                v_descale=None,
            ),
            tag="vLLM Triton",
            metadata=metadata,
            iteration_time_ms=iteration_time_ms,
            warmup_time_ms=warmup_time_ms,
        )
    else:
        vllm_triton_result = None

    conch_result = benchmark_it(
        lambda: varlen_attention(
            query=query,
            key_cache=key_cache,
            value_cache=value_cache,
            cu_seqlens_q=cu_seqlens_q,
            max_seqlen_q=max_seqlen_q,
            seq_lens=seq_lens,
            max_seqlen_k=max_seqlen_k,
            block_table=block_table,
            output=output_conch,
            scale=scale,
            causal=causal,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    if baseline_result is not None:
        baseline_result.print_results(csv=csv)
    if vllm_triton_result is not None:
        vllm_triton_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of bitsandbytes dequantization."""

from enum import Enum

import torch
import triton
import triton.language as tl


class QuantizedDType(Enum):
    """Data types for mapping between Triton/Torch."""

    GENERAL_8BIT = 0
    NF4 = 1
    FP4 = 2


# Triton can only access global values instantiated as tl.constexpr
_QUANTIZED_DTYPE_GENERAL_8BIT: tl.constexpr = tl.constexpr(QuantizedDType.GENERAL_8BIT.value)
_QUANTIZED_DTYPE_NF4: tl.constexpr = tl.constexpr(QuantizedDType.NF4.value)
_QUANTIZED_DTYPE_FP4: tl.constexpr = tl.constexpr(QuantizedDType.FP4.value)


def _get_quantized_dtype(quant_type: str) -> QuantizedDType:
    """Map string quant_type to enum."""
    if quant_type == "nf4":
        return QuantizedDType.NF4
    if quant_type == "fp4":
        return QuantizedDType.FP4
    return QuantizedDType.GENERAL_8BIT


@triton.jit  # type: ignore[misc]
def _nf4_dequantize(x: tl.tensor) -> float:  # noqa: C901, PLR0911, PLR0912
    """Dequantize an NF4 value to floating point."""
    # Note: Triton does not support match/case
    if x == 15:  # noqa: PLR2004
        return 1.0
    if x == 14:  # noqa: PLR2004
        return 0.7229568362236023
    if x == 13:  # noqa: PLR2004
        return 0.5626170039176941
    if x == 12:  # noqa: PLR2004
        return 0.44070982933044434
    if x == 11:  # noqa: PLR2004
        return 0.33791524171829224
    if x == 10:  # noqa: PLR2004
        return 0.24611230194568634
    if x == 9:  # noqa: PLR2004
        return 0.16093020141124725
    if x == 8:  # noqa: PLR2004
        return 0.07958029955625534
    if x == 7:  # noqa: PLR2004
        return 0.0
    if x == 6:  # noqa: PLR2004
        return -0.09105003625154495
    if x == 5:  # noqa: PLR2004
        return -0.18477343022823334
    if x == 4:  # noqa: PLR2004
        return -0.28444138169288635
    if x == 3:  # noqa: PLR2004
        return -0.39491748809814453
    if x == 2:  # noqa: PLR2004
        return -0.5250730514526367
    if x == 1:
        return -0.6961928009986877
    return -1.0


@triton.jit  # type: ignore[misc]
def _fp4_dequantize(x: tl.tensor) -> float:  # noqa: C901, PLR0911, PLR0912
    """Dequantize an FP4 value to floating point."""
    # Note: Triton does not support match/case
    if x == 15:  # noqa: PLR2004
        return -0.25
    if x == 14:  # noqa: PLR2004
        return -0.166666
    if x == 13:  # noqa: PLR2004
        return -0.5
    if x == 12:  # noqa: PLR2004
        return -0.333333
    if x == 11:  # noqa: PLR2004
        return -1.0
    if x == 10:  # noqa: PLR2004
        return -0.666666
    if x == 9:  # noqa: PLR2004
        return -0.0052083333
    if x == 8:  # noqa: PLR2004
        return -0.0
    if x == 7:  # noqa: PLR2004
        return 0.25
    if x == 6:  # noqa: PLR2004
        return 0.166666
    if x == 5:  # noqa: PLR2004
        return 0.5
    if x == 4:  # noqa: PLR2004
        return 0.333333
    if x == 3:  # noqa: PLR2004
        return 1.0
    if x == 2:  # noqa: PLR2004
        return 0.6666666
    if x == 1:
        return 0.0052083333
    return 0.0


@triton.jit  # type: ignore[misc]
def _dequantize_blockwise_kernel(  # noqa: PLR0913
    x_ptr: tl.tensor,
    absmax_ptr: tl.tensor,
    out_ptr: tl.tensor,
    code_ptr: tl.tensor,
    output_size: int,
    cxpr_blocksize: tl.constexpr,
    cxpr_input_blocksize: tl.constexpr,
    cxpr_quantized_dtype: tl.constexpr,
) -> None:
    """Dequantize blockwise kernel."""
    block_index = tl.program_id(0)

    input_block_offset = block_index * cxpr_input_blocksize

    # Get scaling factor for this block
    local_absmax = tl.load(absmax_ptr + block_index)

    # Don't read extra elements if input_size is not perfectly divisible by blocksize
    output_block_offset = block_index * cxpr_blocksize
    this_output_block_size = min(cxpr_blocksize, output_size - output_block_offset)
    this_input_block_size = (
        this_output_block_size if cxpr_quantized_dtype == _QUANTIZED_DTYPE_GENERAL_8BIT else this_output_block_size // 2
    )

    for element_index in range(this_input_block_size):
        if cxpr_quantized_dtype == _QUANTIZED_DTYPE_GENERAL_8BIT:
            x = tl.load(x_ptr + input_block_offset + element_index).to(tl.uint32)
            result = tl.load(code_ptr + x).to(tl.float32) * local_absmax
            tl.store(out_ptr + output_block_offset + element_index, result)
        else:
            # Load packed element
            packed = tl.load(x_ptr + input_block_offset + element_index)

            # Unpack, dequantize, and apply scaling
            if cxpr_quantized_dtype == _QUANTIZED_DTYPE_FP4:
                x1 = _fp4_dequantize(packed >> 4) * local_absmax
                x2 = _fp4_dequantize(packed & 0x0F) * local_absmax
            else:
                x1 = _nf4_dequantize(packed >> 4) * local_absmax
                x2 = _nf4_dequantize(packed & 0x0F) * local_absmax

            # We dequantize two values with each iteration of this loop
            output_element_offset = element_index * 2

            # Store new values
            tl.store(out_ptr + output_block_offset + output_element_offset, x1)
            tl.store(out_ptr + output_block_offset + output_element_offset + 1, x2)


def dequantize_blockwise_launcher(  # noqa: PLR0913
    x: torch.Tensor,
    absmax: torch.Tensor,
    out: torch.Tensor,
    code: torch.Tensor | None,
    blocksize: int,
    output_size: int,
    quant_type: str,
) -> torch.Tensor:
    """Convert FP32/FP16/BF16 input tensor to NF4/FP4."""
    # Blocksize must be a power-of-two
    assert blocksize == triton.next_power_of_2(blocksize)  # noqa: S101
    if quant_type in ("fp4", "nf4"):
        # Only support even output size if unpacking one element into two
        assert output_size % 2 == 0  # noqa: S101

    quantized_dtype = _get_quantized_dtype(quant_type)
    if quantized_dtype == QuantizedDType.GENERAL_8BIT:
        assert code is not None  # noqa: S101

    # If 4-bit, we're unpacking one element from the input into two elements of the output
    input_blocksize = blocksize if quantized_dtype == QuantizedDType.GENERAL_8BIT else blocksize // 2

    # How many blocks are we processing?
    num_blocks = triton.cdiv(output_size, blocksize)

    # Process each block separately
    grid = (num_blocks,)

    # Launch kernel
    _dequantize_blockwise_kernel[grid](
        # Input must be treated as uint8
        # See: https://huggingface.co/docs/bitsandbytes/main/en/fsdp_qlora#quantized-data-storage
        x_ptr=x.view(torch.uint8),
        absmax_ptr=absmax,
        out_ptr=out,
        code_ptr=code,
        output_size=output_size,
        cxpr_blocksize=blocksize,
        cxpr_input_blocksize=input_blocksize,
        cxpr_quantized_dtype=quantized_dtype.value,
    )

    return out



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch copy_blocks benchmark."""

import random
import sys
from typing import Final

import click
import torch

from conch.ops.vllm.copy_blocks import copy_blocks as copy_blocks_conch
from conch.platforms import current_platform
from conch.reference.vllm.copy_blocks import copy_blocks as copy_blocks_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--head-dim",
    required=True,
    type=int,
    default=256,
    help="Head dimension",
)
@click.option(
    "--num-layers",
    required=True,
    type=int,
    default=8,
    help="Number of layers",
)
@click.option(
    "--cache-block-size",
    required=True,
    type=int,
    default=32,
    help="Number of KV vectors in each cache block",
)
@click.option(
    "--num-kv-heads",
    required=False,
    type=int,
    default=8,
    help="Number of kv heads",
)
@click.option(
    "--num-blocks",
    required=False,
    type=int,
    default=2000,
    help="Number of blocks in cache",
)
@click.option(
    "--num-mappings",
    required=False,
    type=int,
    default=512,
    help="Number of mappings to copy",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    head_dim: int,
    num_layers: int,
    cache_block_size: int,
    num_kv_heads: int,
    num_blocks: int,
    num_mappings: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch copy_blocks operation.

    Args:
        head_dim: Head dimension of input tensors.
        num_layers: Sequence length of input tensors.
        cache_block_size: Number of KV vectors in each cache block.
        num_kv_heads: Number of attention kv heads.
        num_blocks: Number of blocks in the cache.
        num_mappings: Number of pairs of blocks to copy.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag to indicate whether or not to print results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = torch.float16

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_dim": head_dim,
            "num_layers": num_layers,
            "cache_block_size": cache_block_size,
            "num_kv_heads": num_kv_heads,
            "num_blocks": num_blocks,
            "num_mappings": num_mappings,
        },
    )

    # Generate random block mappings where each source block is mapped to two
    # destination blocks.
    assert 2 * num_mappings <= num_blocks
    src_blocks = random.sample(range(num_blocks), num_mappings)
    remainig_blocks = list(set(range(num_blocks)) - set(src_blocks))
    dst_blocks = random.sample(remainig_blocks, 2 * num_mappings)
    block_mapping: list[tuple[int, int]] = []
    for i in range(num_mappings):
        src = src_blocks[i]
        dst1 = dst_blocks[2 * i]
        dst2 = dst_blocks[2 * i + 1]
        block_mapping.append((src, dst1))
        block_mapping.append((src, dst2))

    # Create the KV caches.
    key_caches = [
        torch.randn((num_blocks, cache_block_size * num_kv_heads * head_dim), dtype=dtype, device=device)
        for _ in range(num_layers)
    ]
    value_caches = [
        torch.randn((num_blocks, cache_block_size * num_kv_heads * head_dim), dtype=dtype, device=device)
        for _ in range(num_layers)
    ]

    # Clone the KV caches (must happen before calling either kernel bc modifications happen inplace).
    cloned_key_caches = [key_cache.clone() for key_cache in key_caches]
    cloned_value_caches = [value_cache.clone() for value_cache in value_caches]

    # Convert mapping list to tensor
    block_mapping_tensor = torch.tensor(block_mapping, dtype=torch.int64, device=device).view(-1, 2)

    # Run the reference implementation.
    copy_blocks_reference(cloned_key_caches, cloned_value_caches, block_mapping)
    # Call Conch kernel
    copy_blocks_conch(key_caches, value_caches, block_mapping_tensor)

    # Compare the results.
    num_key_matched = 0
    num_value_matched = 0

    for key_cache, cloned_key_cache in zip(key_caches, cloned_key_caches, strict=False):
        if not torch.allclose(key_cache, cloned_key_cache, atol=absolute_tolerance):
            print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
            print(f"Output max diff: {(cloned_key_cache - key_cache).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Reference output: {key_cache}", file=sys.stderr)
                print(f"Conch output: {cloned_key_cache}", file=sys.stderr)
        else:
            num_key_matched += 1

    for value_cache, cloned_value_cache in zip(value_caches, cloned_value_caches, strict=False):
        if not torch.allclose(value_cache, cloned_value_cache, atol=absolute_tolerance):
            print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
            print(f"Output max diff: {(cloned_value_cache - value_cache).abs().max().item()}", file=sys.stderr)

            if verbose:
                print(f"Reference output: {value_cache}", file=sys.stderr)
                print(f"Conch output: {cloned_value_cache}", file=sys.stderr)
        else:
            num_value_matched += 1

    if num_key_matched == num_layers and num_value_matched == num_layers:
        print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

    # Benchmark Reference vs. Conch implementations
    baseline_result = benchmark_it(
        lambda: copy_blocks_reference(
            cloned_key_caches,
            cloned_value_caches,
            block_mapping,
        ),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: copy_blocks_conch(
            key_caches,
            value_caches,
            block_mapping_tensor,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for Triton copy_blocks."""

import random
from typing import Final

import pytest
import torch

from conch.ops.vllm.copy_blocks import copy_blocks as copy_blocks_triton
from conch.platforms import current_platform
from conch.reference.vllm.copy_blocks import copy_blocks as copy_blocks_reference
from conch.third_party.vllm.utils import seed_everything

_DTYPES: Final = [torch.half, torch.bfloat16, torch.float, torch.uint8]
_NUM_LAYERS: Final = [4]
_NUM_HEADS: Final = [1, 4]
_HEAD_SIZES: Final = [128]
_BLOCK_SIZES: Final = [32]
_NUM_BLOCKS: Final = [1000]
_NUM_MAPPINGS: Final = [256]


@pytest.mark.parametrize("num_mappings", _NUM_MAPPINGS)
@pytest.mark.parametrize("num_layers", _NUM_LAYERS)
@pytest.mark.parametrize("num_heads", _NUM_HEADS)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize("block_size", _BLOCK_SIZES)
@pytest.mark.parametrize("num_blocks", _NUM_BLOCKS)
@pytest.mark.parametrize("dtype", _DTYPES)
@torch.inference_mode()
def test_copy_blocks(
    num_mappings: int,
    num_layers: int,
    num_heads: int,
    head_size: int,
    block_size: int,
    num_blocks: int,
    dtype: torch.dtype,
) -> None:
    """Test Triton copy_blocks vs. reference implementations."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    # Generate random block mappings where each source block is mapped to two
    # destination blocks.
    assert 2 * num_mappings <= num_blocks
    src_blocks = random.sample(range(num_blocks), num_mappings)
    remainig_blocks = list(set(range(num_blocks)) - set(src_blocks))
    dst_blocks = random.sample(remainig_blocks, 2 * num_mappings)
    block_mapping: list[tuple[int, int]] = []
    for i in range(num_mappings):
        src = src_blocks[i]
        dst1 = dst_blocks[2 * i]
        dst2 = dst_blocks[2 * i + 1]
        block_mapping.append((src, dst1))
        block_mapping.append((src, dst2))

    # Create the KV caches.
    if dtype == torch.uint8:
        key_caches = [
            torch.randint(0, 255, (num_blocks, block_size * num_heads * head_size), dtype=dtype, device=device)
            for _ in range(num_layers)
        ]
        value_caches = [
            torch.randint(0, 255, (num_blocks, block_size * num_heads * head_size), dtype=dtype, device=device)
            for _ in range(num_layers)
        ]
    else:
        key_caches = [
            torch.randn((num_blocks, block_size * num_heads * head_size), dtype=dtype, device=device)
            for _ in range(num_layers)
        ]
        value_caches = [
            torch.randn((num_blocks, block_size * num_heads * head_size), dtype=dtype, device=device)
            for _ in range(num_layers)
        ]

    # Clone the KV caches (must happen before calling either kernel bc modifications happen ieplace).
    cloned_key_caches = [key_cache.clone() for key_cache in key_caches]
    cloned_value_caches = [value_cache.clone() for value_cache in value_caches]

    # Run the reference implementation.
    copy_blocks_reference(cloned_key_caches, cloned_value_caches, block_mapping)

    # Convert mapping list to tensor
    block_mapping_tensor = torch.tensor(block_mapping, dtype=torch.int64, device=device).view(-1, 2)

    # Call Triton kernel
    copy_blocks_triton(key_caches, value_caches, block_mapping_tensor)

    # Compare the results.
    for key_cache, cloned_key_cache in zip(key_caches, cloned_key_caches, strict=False):
        torch.testing.assert_close(key_cache, cloned_key_cache)
    for value_cache, cloned_value_cache in zip(value_caches, cloned_value_caches, strict=False):
        torch.testing.assert_close(value_cache, cloned_value_cache)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch reshape_and_cache benchmark."""

import random
import sys
from typing import Final

import click
import torch

from conch.ops.vllm.reshape_and_cache import reshape_and_cache as reshape_and_cache_conch
from conch.platforms import current_platform
from conch.reference.vllm.reshape_and_cache import reshape_and_cache as reshape_and_cache_reference
from conch.third_party.vllm.utils import create_kv_caches_with_random, reshape_vllm_kvcache, seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--head-dim",
    required=True,
    type=int,
    default=256,
    help="Head dimension",
)
@click.option(
    "--num-tokens",
    required=True,
    type=int,
    default=512,
    help="Number of tokens",
)
@click.option(
    "--cache-block-size",
    required=True,
    type=int,
    default=32,
    help="Number of KV vectors in each cache block",
)
@click.option(
    "--num-kv-heads",
    required=False,
    type=int,
    default=8,
    help="Number of kv heads",
)
@click.option(
    "--num-blocks",
    required=False,
    type=int,
    default=2000,
    help="Number of blocks in cache",
)
@click.option(
    "--kv-cache-dtype",
    required=False,
    type=click.Choice(["auto", "fp8", "fp8_e4m3"]),
    default="auto",
    help="Dtype of KV cache",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    head_dim: int,
    num_tokens: int,
    cache_block_size: int,
    num_kv_heads: int,
    num_blocks: int,
    kv_cache_dtype: str,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch reshape_and_cache.

    Args:
        head_dim: Head dimension of input tensors.
        num_tokens: Number of tokens to add key/value vectors for into the cache.
        cache_block_size: Number of KV vectors in each cache block.
        num_kv_heads: Number of attention kv heads.
        num_blocks: Number of blocks in the cache.
        kv_cache_dtype: KV Cache dtype.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    if kv_cache_dtype != "auto" and not current_platform.supports_fp8():
        error_msg = "Cannot use FP8 KV Cache because current platform does not support FP8"
        raise NotImplementedError(error_msg)

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = torch.float16

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_dim": head_dim,
            "num_tokens": num_tokens,
            "cache_block_size": cache_block_size,
            "num_kv_heads": num_kv_heads,
            "num_blocks": num_blocks,
            "kv_cache_dtype": kv_cache_dtype,
        },
    )

    # Create a random slot mapping.
    num_slots = cache_block_size * num_blocks
    slot_mapping_lst = random.sample(range(num_slots), num_tokens)
    slot_mapping = torch.tensor(slot_mapping_lst, dtype=torch.long)

    kv = torch.randn(num_tokens, 2, num_kv_heads, head_dim, dtype=dtype)
    key, value = kv.unbind(dim=1)

    k_scale = torch.full((1,), 2.0, dtype=torch.float32, device=device)
    v_scale = torch.full((1,), 3.0, dtype=torch.float32, device=device)

    # Create the KV caches.
    key_caches_vllm, value_caches_vllm = create_kv_caches_with_random(
        num_blocks,
        cache_block_size,
        1,
        num_kv_heads,
        head_dim,
        kv_cache_dtype,
        dtype,
        seed,
        device,
    )

    key_cache_vllm, value_cache_vllm = key_caches_vllm[0], value_caches_vllm[0]
    key_cache, value_cache = reshape_vllm_kvcache(key_cache_vllm.clone(), value_cache_vllm.clone())

    # Run the reference implementation.
    reshape_and_cache_reference(
        key, value, key_cache_vllm, value_cache_vllm, slot_mapping, kv_cache_dtype, k_scale, v_scale
    )

    # Call Conch kernel
    reshape_and_cache_conch(key, value, key_cache, value_cache, slot_mapping, kv_cache_dtype, k_scale, v_scale)

    # Reshape vLLM key/value caches
    key_cache_vllm_out, value_cache_vllm_out = reshape_vllm_kvcache(key_cache_vllm, value_cache_vllm)

    if not torch.allclose(key_cache, key_cache_vllm_out, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(key_cache_vllm_out - key_cache).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {key_cache}", file=sys.stderr)
            print(f"Conch output: {key_cache_vllm_out}", file=sys.stderr)
    else:
        print(f"Key cache matched with atol={absolute_tolerance} :)", file=sys.stderr)

    if not torch.allclose(value_cache, value_cache_vllm_out, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(value_cache_vllm_out - value_cache).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {value_cache}", file=sys.stderr)
            print(f"Conch output: {value_cache_vllm_out}", file=sys.stderr)
    else:
        print(f"Value cache matched with atol={absolute_tolerance} :)", file=sys.stderr)

    # Benchmark Reference vs. Conch implementations
    baseline_result = benchmark_it(
        lambda: reshape_and_cache_reference(
            key,
            value,
            key_cache_vllm,
            value_cache_vllm,
            slot_mapping,
            kv_cache_dtype,
            k_scale,
            v_scale,
        ),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: reshape_and_cache_conch(
            key,
            value,
            key_cache,
            value_cache,
            slot_mapping,
            kv_cache_dtype,
            k_scale,
            v_scale,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of vLLM reshape_and_cache_kernel."""

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _reshape_and_cache_kernel(
    # Pointers to tensors
    key_ptr: tl.tensor,
    value_ptr: tl.tensor,
    key_cache_ptr: tl.tensor,
    value_cache_ptr: tl.tensor,
    slot_mapping_ptr: tl.tensor,
    k_scale_ptr: tl.tensor,
    v_scale_ptr: tl.tensor,
    # Strides of relevant tensors
    k_token_stride: int,
    k_head_stride: int,
    k_head_element_stride: tl.constexpr,
    v_token_stride: int,
    v_head_stride: int,
    v_head_element_stride: tl.constexpr,
    kv_cache_page_stride: int,
    kv_cache_block_stride: int,
    kv_cache_head_stride: int,
    kv_cache_head_element_stride: tl.constexpr,
    # Scalars
    cache_block_size: int,
    # Constexprs
    cxpr_head_size: tl.constexpr,
    cxpr_apply_fp8_scaling: tl.constexpr,
) -> None:
    """Implementation of reshape_and_cache kernel.

    Args:
        key_ptr: Pointer to tensor of new key vectors, shape: (num_tokens, num_kv_heads, head_size).
        value_ptr: Pointer to tensor of new value vectors, shape: (num_tokens, num_kv_heads, head_size).
        key_cache_ptr: Pointer to tensor of key cache, shape: (num_pages, cache_block_size, num_kv_heads, head_size).
        value_cache_ptr: Pointer to tensor of value cache, shape: (num_pages, cache_block_size, num_kv_heads, head_size).
        slot_mapping_ptr: Pointer to slot mapping tensor, shape: (num_tokens,).
        k_scale_ptr: Pointer to Fp8 scaling factor for k.
        v_scale_ptr: Pointer to Fp8 scaling factor for v.
        k_token_stride: Stride of key tensor in 0th dimension.
        k_head_stride: Stride of key tensor in 1st dimension.
        k_head_element_stride: Stride of key tensor in 2nd dimension.
        v_token_stride: Stride of value tensor in 0th dimension.
        v_head_stride: Stride of value tensor in 1st dimension.
        v_head_element_stride: Stride of value tensor in 2nd dimension.
        kv_cache_page_stride: Stride of key/value cache tensors in 0th dimension.
        kv_cache_block_stride: Stride of key/value cache tensors in 1st dimension.
        kv_cache_head_stride: Stride of key/value cache tensors in 2nd dimension.
        kv_cache_head_element_stride: Stride of key/value cache tensors in 3rd dimension.
        cache_block_size: Size of each cache block / page in the KV cache.
        cxpr_head_size: Head size / dimension for the attention head (must be power of two!).
        cxpr_apply_fp8_scaling: Whether or not to apply FP8 scaling.
    """
    # What token is this program processing?
    token_index = tl.program_id(0)

    # Get index of slot for this token from mapping tensor
    slot_index = tl.load(slot_mapping_ptr + token_index)

    # If slot index is negative its a padding token that should be ignored
    if slot_index < 0:
        return

    # What head is this program processing?
    head_index = tl.program_id(1)

    # Calculate offset into key/value tensors to get to the token for this program
    k_token_offset = token_index * k_token_stride
    v_token_offset = token_index * v_token_stride
    # Calculate offset into key/value tensors to get to the head for this program
    k_head_offset = head_index * k_head_stride
    v_head_offset = head_index * v_head_stride
    # Offsets for each element of the head
    k_head_offsets = tl.arange(0, cxpr_head_size) * k_head_element_stride
    v_head_offsets = tl.arange(0, cxpr_head_size) * v_head_element_stride

    # Load key/value vectors for this token/head
    key = tl.load(key_ptr + k_token_offset + k_head_offset + k_head_offsets)
    value = tl.load(value_ptr + v_token_offset + v_head_offset + v_head_offsets)

    # Apply FP8 scaling if necessary
    if cxpr_apply_fp8_scaling:
        k_scale = tl.load(k_scale_ptr)
        k_scale = 1.0 / k_scale
        key *= k_scale

        v_scale = tl.load(v_scale_ptr)
        v_scale = 1.0 / v_scale
        value *= v_scale

        key = key.to(key_cache_ptr.dtype.element_ty)
        value = value.to(value_cache_ptr.dtype.element_ty)

    # Calculate index of page (value in range(0, num_pages))
    page_index = slot_index // cache_block_size
    # Calculate entry index inside of a cache block/page for this slot (value in range(0, cache_block_size))
    entry_index = slot_index % cache_block_size

    # Calculate offset into key/value cache tensors to get to the cache block we're copying into
    kv_page_offset = page_index * kv_cache_page_stride
    # Calculate offset in a cache block to get to the entry for we're copying into
    kv_cache_entry_offset = entry_index * kv_cache_block_stride
    # Calculate offset into key/value cache tensors to get to the head we're copying into
    kv_cache_head_offset = head_index * kv_cache_head_stride

    # Offsets for each element of the head
    kv_head_offsets = tl.arange(0, cxpr_head_size) * kv_cache_head_element_stride

    # Store key/value vectors into cache
    tl.store(key_cache_ptr + kv_page_offset + kv_cache_entry_offset + kv_cache_head_offset + kv_head_offsets, key)
    tl.store(value_cache_ptr + kv_page_offset + kv_cache_entry_offset + kv_cache_head_offset + kv_head_offsets, value)


def reshape_and_cache_launcher(
    key: torch.Tensor,
    value: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    slot_mapping: torch.Tensor,
    kv_cache_dtype: str = "auto",
    k_scale: torch.Tensor | None = None,
    v_scale: torch.Tensor | None = None,
    strict: bool = False,
) -> None:
    """Launch reshape_and_cache kernel.

    Args:
        key: New key vectors, shape: (num_tokens, num_kv_heads, head_size).
        value: New value vectors, shape: (num_tokens, num_kv_heads, head_size).
        key_cache: Key cache, shape: (num_pages, cache_block_size, num_kv_heads, head_size).
        value_cache: Value cache, shape: (num_pages, cache_block_size, num_kv_heads, head_size).
        slot_mapping: Tensor listing what slots in the cache each key/value vector should be placed in, shape: (num_tokens,).
        kv_cache_dtype: String datatype of kv cache elements.
        k_scale: Fp8 scaling factor for k.
        v_scale: Fp8 scaling factor for v.
    """
    # Assume sizes already checked if calling launcher. For interface with strict size checking, call `ops.reshape_and_cache()` with `strict=True`.
    _, num_kv_heads, head_size = key.shape
    num_pages, cache_block_size, _, _ = key_cache.shape

    # Note: In vLLM v1, slot_mapping is the only tensor that can be trusted to tell the correct number of tokens
    num_tokens = slot_mapping.size(0)

    if strict:
        assert key.shape == value.shape
        assert key_cache.shape == value_cache.shape

        assert key_cache.stride(0) == value_cache.stride(0)
        assert key_cache.stride(1) == value_cache.stride(1)
        assert key_cache.stride(2) == value_cache.stride(2)
        assert key_cache.stride(3) == value_cache.stride(3)
        assert key_cache.stride(3) == 1

        assert cache_block_size == triton.next_power_of_2(cache_block_size), "Cache block size must be a power of two!"
        assert head_size == triton.next_power_of_2(head_size), "Head size must be a power of two!"

    apply_fp8_scaling: tl.constexpr = "fp8" in kv_cache_dtype

    if strict and apply_fp8_scaling:
        assert k_scale is not None
        assert v_scale is not None
        assert k_scale.numel() == 1
        assert v_scale.numel() == 1

    # Parallelize over the number of tokens and number of kv heads
    grid = (num_tokens, num_kv_heads)

    # Launch kernel
    _reshape_and_cache_kernel[grid](
        # Tensors
        key,
        value,
        key_cache,
        value_cache,
        slot_mapping,
        k_scale,
        v_scale,
        # Strides of relevant tensors
        key.stride(0),
        key.stride(1),
        key.stride(2),
        value.stride(0),
        value.stride(1),
        value.stride(2),
        key_cache.stride(0),
        key_cache.stride(1),
        key_cache.stride(2),
        key_cache.stride(3),
        # Scalars
        cache_block_size,
        # Constexprs
        head_size,
        apply_fp8_scaling,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for mixed precision GEMM."""

import math
from typing import Final

import pytest
import torch

from conch.ops.quantization.gemm import mixed_precision_gemm
from conch.platforms import current_platform
from conch.third_party.vllm.quant_utils import pack_rows, quantize_weights
from conch.third_party.vllm.scalar_type import ScalarType, scalar_types
from conch.third_party.vllm.utils import seed_everything


def _quantize_and_pack(
    w: torch.Tensor,
    wtype: ScalarType,
    group_size: int,
    use_zero_points: bool,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor | None, torch.Tensor | None]:
    """Quantize and pack."""

    w_ref, w_q, w_s, w_zp = quantize_weights(
        w, wtype, group_size, zero_points=use_zero_points, ref_zero_points_after_scales=False
    )

    w_q_packed = pack_rows(w_q, wtype.size_bits, *w_q.shape)

    return w_ref, w_q_packed, w_s, w_zp


@pytest.mark.parametrize(("m_dim", "k_dim", "n_dim"), [(128, 256, 128), (1024, 1024, 1024), (4096, 2048, 4096)])
@pytest.mark.parametrize(
    "weight_dtype", [scalar_types.uint4b8, scalar_types.uint8b128, scalar_types.uint4, scalar_types.uint8]
)
@pytest.mark.parametrize("use_zero_points", [True, False])
@pytest.mark.parametrize("input_dtype", [torch.float16, torch.bfloat16])
def test_gemm(
    m_dim: int,
    k_dim: int,
    n_dim: int,
    weight_dtype: ScalarType,
    use_zero_points: bool,
    input_dtype: torch.dtype,
) -> None:
    """Test mixed precision GEMM Triton kernel."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    group_size: Final = 128
    assert group_size <= k_dim

    a = (10 * (torch.rand((m_dim, k_dim), dtype=torch.float32, device=device) - 0.3)).to(input_dtype)
    b = (10 * (torch.rand((k_dim, n_dim), dtype=torch.float32, device=device) - 0.3)).to(input_dtype)

    w_ref, w_q_packed, w_s, w_zp = _quantize_and_pack(b, weight_dtype, group_size, use_zero_points)

    # For mypy
    if use_zero_points:
        assert w_zp is not None
    assert w_s is not None

    output_ref = torch.matmul(a, w_ref)

    triton_output = mixed_precision_gemm(
        a, w_q_packed, w_s, w_zp, weight_dtype.size_bits, weight_dtype.bias, group_size
    )

    atol = min(5e-2 * math.sqrt(k_dim), 1)
    torch.testing.assert_close(triton_output, output_ref, rtol=1e-1, atol=atol)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of FP8 quantization kernels."""

import torch

from conch import envs
from conch.platforms import current_platform


def _scaled_fp8_quant_pytorch_ref(x: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
    """PyTorch reference fp8 quant impl."""
    quant_dtype = torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn
    finfo = torch.finfo(quant_dtype)
    scale = scale.reciprocal()
    qweight = (x.to(torch.float32) * scale).clamp(min=finfo.min, max=finfo.max)
    return qweight.to(quant_dtype)


def _scaled_fp8_quant_vllm_ref(x: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
    """vLLM reference fp8 quant impl."""
    from vllm._custom_ops import scaled_fp8_quant as scaled_fp8_quant_vllm

    output, _ = scaled_fp8_quant_vllm(x, scale)
    return output  # type: ignore[no-any-return, unused-ignore]


def scaled_fp8_quant(
    input_tensor: torch.Tensor,
    scale: torch.Tensor | None = None,
) -> torch.Tensor:
    """Scaled fp8 quantization (currently only static supported)."""
    if scale is None:
        error_msg = "Dynamic quantization not implemented"
        raise NotImplementedError(error_msg)

    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _scaled_fp8_quant_vllm_ref(input_tensor, scale)

    return _scaled_fp8_quant_pytorch_ref(input_tensor, scale)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of [SiLU](https://pytorch.org/docs/stable/generated/torch.nn.SiLU.html) and mul."""

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _silu_and_mul_kernel(  # noqa: PLR0913
    # Pointers to tensors
    output_ptr: tl.tensor,  # [..., d]
    x_ptr: tl.tensor,  # [..., 2 * d]
    # Strides of relevant tensors
    output_stride: int,
    input_stride: int,
    # Scalars
    d: int,
    # Constexprs
    cxpr_block_size: tl.constexpr,
) -> None:
    """Implementation of SiLU and multiply kernel.

    Args:
        output_ptr: Pointer to output tensor, shape: (num_tokens, d) or (batch_size, sequence_length, d).
        x_ptr: Pointer to input tensor, shape: (num_tokens, 2 * d) or (batch_size, sequence_length, 2 * d).
        output_stride: Stride of output tensor between elements in the last dimension.
        input_stride: Stride of input tensor between elements in the last dimension.
        d: Size of last dimension of output.
        cxpr_block_size: Number of elements to process at once.
    """
    token_index = tl.program_id(0)

    x_token_offset = token_index * input_stride
    y_token_offset = x_token_offset + d
    output_token_offset = token_index * output_stride

    block_offsets = tl.arange(0, cxpr_block_size)

    for _ in tl.range(0, d, cxpr_block_size):
        mask = block_offsets < d

        # For parity with vLLM, compute `x * sigmoid(x)` in fp32
        x = tl.load(x_ptr + x_token_offset + block_offsets, mask=mask, other=0.0).to(tl.float32)
        y = tl.load(x_ptr + y_token_offset + block_offsets, mask=mask, other=0.0)

        silu = (x * tl.sigmoid(x)).to(x_ptr.dtype.element_ty)
        silu *= y

        tl.store(output_ptr + output_token_offset + block_offsets, silu, mask=mask)

        block_offsets += cxpr_block_size


def silu_and_mul_launcher(
    output: torch.Tensor,
    x: torch.Tensor,
) -> None:
    """Launch silu_and_mul kernel.

    Args:
        x: Input tensor, of shape (num_tokens, 2 * d) or (batch_size, seq_len, 2 * d).
        output: Output tensor, of shape (num_tokens, d) or (batch_size, seq_len, d).
    """
    d = x.size(-1) // 2
    cxpr_block_size = min(1024, triton.next_power_of_2(d))

    num_tokens = x.numel() // x.size(-1)

    assert output.shape == x.shape[:-1] + (d,), "Output shape must match input shape with last dimension halved!"  # noqa: S101
    assert x.is_contiguous()  # noqa: S101
    assert output.is_contiguous()  # noqa: S101

    # Parallelize over the number of tokens
    grid = (num_tokens,)

    # Launch kernel
    _silu_and_mul_kernel[grid](
        # Tensors
        output_ptr=output,
        x_ptr=x,
        # Strides of relevant tensors
        output_stride=output.stride(-2),
        input_stride=x.stride(-2),
        # Scalars
        d=d,
        # Constexprs
        cxpr_block_size=cxpr_block_size,
    )



# SPDX-License-Identifier: Apache-2.0

import numpy
import torch

from conch.third_party.vllm.scalar_type import ScalarType


def get_pack_factor(num_bits):
    assert 32 % num_bits == 0, f"Unsupported num_bits = {num_bits}"
    return 32 // num_bits


def quantize_weights(
    w: torch.Tensor,
    quant_type: ScalarType,
    group_size: int | None,
    zero_points: bool = False,
    ref_zero_points_after_scales: bool = False,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor | None, torch.Tensor | None]:
    assert quant_type.is_integer(), "Floating point quantization may work but has not been tested"
    assert not zero_points or group_size is not None, (
        "to have group zero points, group_size must be provided " "(-1 group_size is channelwise)"
    )

    orig_device = w.device
    orig_type = w.dtype
    size_k, size_n = w.shape

    assert w.is_floating_point(), "w must be float"

    if group_size == -1:
        group_size = size_k

    # Reshape to [groupsize, -1]
    if group_size is not None and group_size < size_k:
        w = w.reshape((-1, group_size, size_n))
        w = w.permute(1, 0, 2)
        w = w.reshape((group_size, -1))

    # Compute scale for each group
    max_val = torch.max(w, 0, keepdim=True).values
    min_val = torch.min(w, 0, keepdim=True).values

    max_q_val = quant_type.max()
    min_q_val = quant_type.min()

    w_s = torch.Tensor([1.0]).to(w.device)  # unscaled case
    maybe_w_zp = None
    if group_size is not None:
        if zero_points:
            assert not quant_type.is_signed() and quant_type.max() > 0
            w_s = (max_val - min_val).clamp(min=1e-5) / quant_type.max()
            maybe_w_zp = torch.round(torch.abs(min_val / w_s)).clamp(min_q_val, max_q_val).int()
        else:
            # If the bias is such that there are no possible negative/positive
            #  values, set the max value to inf to avoid divide by 0
            w_s = torch.max(
                abs(max_val / (max_q_val if max_q_val != 0 else torch.inf)),
                abs(min_val / (min_q_val if min_q_val != 0 else torch.inf)),
            )

    # Quantize
    w_q = torch.round(w / w_s).int() + (maybe_w_zp if zero_points else 0)
    w_q = torch.clamp(w_q, min_q_val, max_q_val)

    # Compute ref (dequantized)
    # For some kernels (namely Machete) the zero-points are applied after the
    # scales are applied, for this case computing the reference in similar way
    # allows us to use tighter error tolerances in our unit tests.
    if ref_zero_points_after_scales and maybe_w_zp is not None:
        w_ref = w_q.to(orig_type) * w_s - maybe_w_zp.to(orig_type) * w_s
    else:
        w_ref = (w_q - (maybe_w_zp if zero_points else 0)).to(orig_type) * w_s

    if quant_type.has_bias():
        w_q += quant_type.bias

    # Restore original shapes
    if group_size is not None and group_size < size_k:

        def reshape_w(w):
            w = w.reshape((group_size, -1, size_n))
            w = w.permute(1, 0, 2)
            w = w.reshape((size_k, size_n)).contiguous()
            return w

        w_q = reshape_w(w_q)
        w_ref = reshape_w(w_ref)
        w_s = w_s.reshape((-1, size_n)).contiguous()

    if maybe_w_zp is not None:
        maybe_w_zp = maybe_w_zp.reshape((-1, size_n)).contiguous()
        maybe_w_zp = maybe_w_zp.to(device=orig_device)

    return (
        w_ref.to(device=orig_device),
        w_q.to(device=orig_device),
        w_s if group_size is not None else None,
        maybe_w_zp,
    )


def pack_rows(
    q_w: torch.Tensor,
    num_bits: int,
    size_k: int,
    size_n: int,
) -> torch.Tensor:
    assert q_w.shape == (size_k, size_n)

    pack_factor = get_pack_factor(num_bits)
    assert size_k % pack_factor == 0

    orig_device = q_w.device

    q_w = q_w.cpu().numpy().astype(numpy.uint32)

    q_res = numpy.zeros((size_k // pack_factor, size_n), dtype=numpy.uint32)

    for i in range(pack_factor):
        q_res |= q_w[i::pack_factor, :] << num_bits * i

    q_res = torch.from_numpy(q_res.astype(numpy.int32)).to(orig_device)
    return q_res



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of RMS norm."""

import torch
import triton
import triton.language as tl


@triton.jit  # type: ignore[misc]
def _rms_norm_kernel(  # noqa: PLR0913
    # Tensors
    output_ptr: tl.tensor,  # [..., hidden_size]
    x_ptr: tl.tensor,  # [..., hidden_size]
    residual_ptr: tl.tensor,  # [..., hidden_size]
    weight_ptr: tl.tensor,  # [hidden_size]
    # Scalars
    hidden_size: int,
    epsilon: float,
    # Constexprs
    cxpr_block_size: tl.constexpr,
    cxpr_use_residual: tl.constexpr,
) -> None:
    """Implementation of RMS norm kernel.

    Args:
        output_ptr: Pointer to output tensor, shape: (num_tokens, hidden_size).
        x_ptr: Pointer to input tensor, shape: (num_tokens, hidden_size).
        residual_ptr: Pointer to residual tensor, shape: (num_tokens, hidden_size).
        weight_ptr: Pointer to weight tensor, shape: (hidden_size,).
        hidden_size: Hidden size.
        epsilon: Epsilon value.
        cxpr_block_size: Number of elements to process at once.
        cxpr_use_residual: Whether to use residual tensor.
    """
    token_index = tl.program_id(0)
    token_offset = token_index * hidden_size

    block_offsets = tl.arange(0, cxpr_block_size)
    mask = block_offsets < hidden_size

    x = tl.load(x_ptr + token_offset + block_offsets, mask=mask)
    w = tl.load(weight_ptr + block_offsets, mask=mask)

    if cxpr_use_residual:
        # Load residual, add it to x, and store it
        residual = tl.load(residual_ptr + token_offset + block_offsets, mask=mask)
        x += residual
        tl.store(residual_ptr + token_offset + block_offsets, x, mask=mask)

        # If we are using the residual, we will write the result to the input tensor
        output_ptr = x_ptr

    # For parity with vLLM, we will use fp32 here
    x = x.to(tl.float32)
    mean_of_squares = tl.sum(x * x) / hidden_size
    rms_inv = tl.rsqrt(mean_of_squares + epsilon)

    result = (x * rms_inv).to(x_ptr.dtype.element_ty) * w

    tl.store(output_ptr + token_offset + block_offsets, result, mask=mask)


def rms_norm_launcher(
    output: torch.Tensor,
    x: torch.Tensor,
    weight: torch.Tensor,
    epsilon: float,
) -> None:
    """Launch rms_norm kernel.

    Args:
        output: Output tensor, of shape (..., hidden_size).
        x: Input tensor, of shape (..., hidden_size).
        weight: Weight tensor, of shape (hidden_size,).
        epsilon: Epsilon value.
    """
    hidden_size = x.size(-1)
    num_tokens = x.numel() // hidden_size

    assert output.shape == x.shape, "Output shape must match input shape"  # noqa: S101
    assert output.stride(-2) == x.stride(-2), "Output and input strides must match"  # noqa: S101
    assert output.stride(-2) == hidden_size, "Hidden size must match second-to-last stride of input/output"  # noqa: S101
    assert weight.size(0) == hidden_size, "Weight size must match hidden size"  # noqa: S101

    assert output.is_contiguous()  # noqa: S101
    assert x.is_contiguous()  # noqa: S101
    assert weight.is_contiguous()  # noqa: S101

    # Parallelize over the number of tokens
    grid = (num_tokens,)

    # Launch kernel
    _rms_norm_kernel[grid](
        # Tensors
        output_ptr=output,
        x_ptr=x,
        residual_ptr=None,
        weight_ptr=weight,
        # Scalars
        hidden_size=hidden_size,
        epsilon=epsilon,
        # Constexprs
        # Note: we _could_ run out of shared memory here if hidden_size is too large.
        # If this is a concern, we could allocate some extra memory for the sum of squares and then
        # reduce it in a second kernel.
        cxpr_block_size=triton.next_power_of_2(hidden_size),
        cxpr_use_residual=False,
    )


def fused_add_rms_norm_launcher(
    x: torch.Tensor,
    residual: torch.Tensor,
    weight: torch.Tensor,
    epsilon: float,
) -> None:
    """Launch rms_norm kernel.

    Args:
        x: Input tensor, of shape (..., hidden_size).
        residual: Residual tensor, of shape (..., hidden_size).
        weight: Weight tensor, of shape (hidden_size,).
        epsilon: Epsilon value.
    """
    hidden_size = x.size(-1)
    num_tokens = x.numel() // hidden_size

    assert x.shape == residual.shape, "Input shape must match residual shape"  # noqa: S101
    assert x.stride(-2) == residual.stride(-2), "Input and residual strides must match"  # noqa: S101
    assert x.stride(-2) == hidden_size, "Hidden size must match second-to-last stride of input"  # noqa: S101
    assert weight.size(0) == hidden_size, "Weight size must match hidden size"  # noqa: S101

    assert x.is_contiguous()  # noqa: S101
    assert residual.is_contiguous()  # noqa: S101
    assert weight.is_contiguous()  # noqa: S101

    # Parallelize over the number of tokens
    grid = (num_tokens,)

    # Launch kernel
    _rms_norm_kernel[grid](
        # Tensors
        output_ptr=None,
        x_ptr=x,
        residual_ptr=residual,
        weight_ptr=weight,
        # Scalars
        hidden_size=hidden_size,
        epsilon=epsilon,
        # Constexprs
        # Note: we _could_ run out of shared memory here if hidden_size is too large.
        # If this is a concern, we could allocate some extra memory for the sum of squares and then
        # reduce it in a second kernel.
        cxpr_block_size=triton.next_power_of_2(hidden_size),
        cxpr_use_residual=True,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch rotary_embedding benchmark."""

import sys
from typing import Final

import click
import torch

from conch.ops.embedding.rotary_embedding import rotary_embedding as rotary_embedding_conch
from conch.platforms import current_platform
from conch.reference.embedding.rotary_embedding import compute_cos_sin_cache
from conch.reference.embedding.rotary_embedding import rotary_embedding as rotary_embedding_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--head-size",
    required=False,
    type=int,
    default=256,
    help="Feedforward hidden size",
)
@click.option(
    "--num-heads",
    required=False,
    type=int,
    default=8,
)
@click.option(
    "--num-tokens",
    required=False,
    type=int,
    default=8192,
    help="Number of tokens",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag to output results in CSV format",
)
def main(
    head_size: int,
    num_heads: int,
    num_tokens: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch RotaryEmbedding op.

    Args:
        head_size: Head size.
        num_heads: Number of heads.
        num_tokens: Number of tokens.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    base: Final = 10000
    is_neox_style: Final = True
    rotary_dim: Final = head_size
    max_position = num_tokens

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_size": head_size,
            "num_heads": num_heads,
            "num_tokens": num_tokens,
        },
    )

    cos_sin_cache = compute_cos_sin_cache(base, rotary_dim, max_position)

    positions = torch.randint(0, max_position, (num_tokens,))
    query = torch.randn(num_tokens, num_heads * head_size)
    key = torch.randn_like(query)
    # Need to clone since updates are performed in-place
    query_ref = torch.clone(query)
    key_ref = torch.clone(key)

    query_ref, key_ref = rotary_embedding_reference(
        positions,
        query_ref,
        key_ref,
        cos_sin_cache,
        rotary_dim,
        head_size,
        is_neox_style=is_neox_style,
    )

    query_conch, key_conch = rotary_embedding_conch(
        positions,
        query,
        key,
        head_size,
        cos_sin_cache,
        is_neox=is_neox_style,
    )

    if not torch.allclose(query_ref, query_conch, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(query_conch - query_ref).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {query_ref}", file=sys.stderr)
            print(f"Conch output: {query_conch}", file=sys.stderr)
    else:
        print(f"Query matched with atol={absolute_tolerance} :)", file=sys.stderr)

    if not torch.allclose(key_ref, key_conch, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(key_conch - key_ref).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {key_ref}", file=sys.stderr)
            print(f"Conch output: {key_conch}", file=sys.stderr)
    else:
        print(f"Key matched with atol={absolute_tolerance} :)", file=sys.stderr)

    baseline_result = benchmark_it(
        lambda: rotary_embedding_reference(
            positions,
            query_ref,
            key_ref,
            cos_sin_cache,
            rotary_dim,
            head_size,
            is_neox_style=is_neox_style,
        ),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )
    conch_result = benchmark_it(
        lambda: rotary_embedding_conch(
            positions,
            query,
            key,
            head_size,
            cos_sin_cache,
            is_neox=is_neox_style,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()
